# Started by A.Stehli@KIT 04/2018
# time domain virtual AWG class


# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from typing import Dict, List, Generator, Any
import numpy as np
from ipywidgets import interact, widgets, Layout
import logging

import qkit
import qkit.measure.timedomain.pulse_sequence as ps
import qkit.measure.timedomain.awg.load_tawg as load_tawg

PLOT_ENABLE = False
try:
    import matplotlib.pyplot as plt

    PLOT_ENABLE = True
except ImportError:
    pass

# Helper functions
def all_are_same(array):
    """Check if all elements of a list are the same."""
    return all(item == array[0] for item in array)


def _vars_len(variables: Dict[str, List[float]]):
    if variables:
        return len(list(variables.values())[0])
    else:
        return 0  # No variables in dictionary


def dictify_variable_lists(variables):
    # type: (Dict[List[float]]) -> Generator[Dict[float]]
    for i in range(_vars_len(variables)):
        single_variables = {}
        for key in variables.keys():
            single_variables[key] = variables[key][i]
        yield single_variables


class TdChannel(object):
    """
    Class managing the channels for the VirtualAWG class.
    These channels are similar to the channel of an awg.

    Attributes:
        add_sequence:      add sequence maintaining currently stored sequences
        set_sequence:      add sequence deleting previously stored sequences
        delete_sequence:   delete previously stored sequences
        set_interleave:    turn interleaving of sequences on/off

        get_sequence_dict: returns a dictionary containing all important channel attributes
        plot:              plot the pulse sequences

    TODO:
        Add option to compensate linearly (for Bias Tees with finite decay time).
    """

    def __init__(self, sample, name):
        """
        Inits TdChannel with sample and name:
            sample: sample object
            name:   name of the channel
        """
        self.name = name
        self._sample = sample
        self._sequences = []
        self._variables = []  # type: List[Dict[List[float]]]
        self._interleave = False
        # Dictionary for x-axis scaling
        self._x_unit = {"s": 1, "ms": 1e-3, "us": 1e-6, "ns": 1e-9}

    @property
    def sequence_count(self):
        """The number of loaded sequences."""
        return len(self._sequences)

    def add_sequence(
        self, sequence: ps.PulseSequence, **variables: List[float]
    ) -> bool:
        """
        Append sequence to sequences currently stored in channel.

        Args:
            sequence:     sequence object
            **variables:  One or multiple keywords matching the variable names of time functions and each containing a list of values.
                          In a T1 measurement this would be the wait time between pi-pulse and readout tone.
        """
        # Check that lists are given for all necessary variables
        if sequence.variable_names:
            if not variables or sequence.variable_names != set(variables.keys()):
                logging.error(
                    "Lists for the variables of the sequence have to be specified. Given variables do not match with required ones. "
                    + "The following keyword arguments are required: {}.".format(
                        ", ".join(sequence.variable_names)
                    )
                )
                return False

            # Check if all variable lists have the same length and are non-empty
            if not all_are_same([len(v) for v in variables.values()]):
                logging.error("Length of variable lists do not match.")
                return False
            elif _vars_len(variables) == 0:
                logging.error(
                    "The lists containing values of the variables must not be empty."
                )
                return False

        # Add sequence and variable lists to channel
        self._sequences.append(sequence)
        self._variables.append(variables)

        # Check if interleaving should be done and if so, if it is still possible
        if (len(self._sequences) >= 1) and self._interleave:
            if not self.set_interleave():
                logging.error(
                    "Dimension of new time array does not fit previous inputs.\n"
                    + "Interleaving no longer possible."
                )

        return True

    def set_sequence(self, sequence, **variables):
        """
        Set sequence in channel to sequence deleting previously stored sequences.

        Input:
            sequence:     sequence object
            **variables:  One or multiple keywords matching the variable names of time functions and each containing a list of values.
                          In a T1 measurement this would be the wait time between pi-pulse and readout tone.
        """
        # Delete previously stored sequences
        self.delete_sequence()

        # Add new sequence
        return self.add_sequence(sequence, **variables)

    def delete_sequence(self, seq_nr=None):
        """
        Delete sequence number seq_nr (counting from 0).
        If seq_nr is None, all sequences are deleted.

        Args:
            seq_nr: Number of the sequence to be deleted.
        """
        if seq_nr is None:
            self._sequences = []
            self._variables = []
        else:
            self._sequences.pop(seq_nr)
            self._variables.pop(seq_nr)
        return True

    def get_sequence(self, num: int):
        # type: (int) -> Tuple[PulseSequence, Dict[str, List[Any]]]
        """Returns the PulseSequence object and variables for a given index."""
        return self._sequences[num], self._variables[num]

    def get_sequence_dict(self):
        """
        Returns a dictionary featuring all important channel attributes:
            sequence_i: list of all pulses of sequences number i (counting from 0)
            par_i: parameter values for which sequence is called (usually this parameter is a time, e.g. separating two pulses)
            value of the interleave attribute
        """
        seq_dict = {}
        for i in range(len(self._sequences)):
            seq_dict["sequence_%i" % i] = self._sequences[i].get_pulses()
            seq_dict["par_%i" % i] = self._variables[i]
        if seq_dict:
            seq_dict["interleave"] = self._interleave
        return seq_dict

    def set_interleave(self, use_interleave=True):
        """
        Sequences with an equal number of timesteps may be interleaved.
        In order for this to work they must have the same number of time steps.

        Args:
            use_interleave: if true, the sequences in this channel are interleaved.
        """
        # By default, interleave mode is off
        self._interleave = False

        if not use_interleave:
            # Disabling interleave mode always works
            return True

        if not self._variables:
            # If there are no sequences yet, one can also turn on interleave mode
            self._interleave = True
            return True

        # Check if all variable vectors have same length
        # (Only need to check first variable of each vector, as within we already checked during add procedure)
        if not all_are_same([_vars_len(vs) for vs in self._variables]):
            logging.error(
                "Only sequences that have the same number of variables can be interleaved."
            )
            return False

        self._interleave = True
        return True

    @property
    def is_interleaved(self):
        """If the sequences in this channel are to be played interleaved."""
        return self._interleave

    def plot(self, show_quadrature="", x_unit="ns"):
        """
        Plots the sequences stored in channel.
        A slider provides the option to sweep through different time values.
        Readout pulse is fixed at t = 0, where is t>0 before the readout tone.

        Args:
            x_unit: unit of the x-axis in the plot. Options are "s", "ms", "us", "ns".
        """
        show_iq = show_quadrature in ["I", "Q"]
        sequences, readout_indices = self._get_sequences(IQ_mixing=show_iq)
        seq_max = len(readout_indices) - 1

        # If there is no complex information in there, we just present the real part:
        has_complex_values = any([any(np.iscomplex(seq)) for seq in sequences])
        if show_quadrature == "I" or not has_complex_values:
            sequences = [np.real(seq) for seq in sequences]
        elif show_quadrature == "Q":
            sequences = [np.imag(seq) for seq in sequences]

        bounds = self._get_boundaries(sequences, readout_indices, x_unit)

        interact(
            lambda sequence: self._plot_sequence(
                sequences[sequence], readout_indices[sequence], x_unit, bounds
            ),
            sequence=widgets.IntSlider(
                value=0, min=0, max=seq_max, layout=Layout(width="98%", height="50px")
            ),
        )
        return True

    def _plot_sequence(self, seq, ro_ind, x_unit, bounds, col="C0", plot_readout=True):
        """
        The actual plotting of the sequences happens here.

        Input:
            seq:    sequence (as array) to be plotted
            ro_ind: index of the readout in seq
            x_unit: x_unit for the time axis
            bounds: boundaries for the plot (xmin, xmax, ymin, ymax)
        """
        if not PLOT_ENABLE:
            raise ImportError("matplotlib not found.")

        if plot_readout:
            fig = plt.figure(figsize=(18, 6))
        xmin, xmax, ymin, ymax = bounds
        samplerate = self._sample.clock
        time = -(np.arange(0, len(seq) + 1, 1) - ro_ind) / (
            samplerate * self._x_unit[x_unit]
        )
        # make sure last point of the waveform goes to zero
        seq = np.append(seq, 0)

        try:
            readout_tone = self._sample.rec_pulselength
        except AttributeError:
            readout_tone = self._sample.readout_tone_length

        # plot sequence
        plt.plot(time, seq, col)
        plt.fill(time, seq, color=col, alpha=0.2)
        # plot readout
        if plot_readout:
            plt.fill(
                [
                    0,
                    0,
                    -readout_tone / self._x_unit[x_unit],
                    -readout_tone / self._x_unit[x_unit],
                ],
                [0, ymax, ymax, 0],
                color="C7",
                alpha=0.3,
            )
            # add label for readout
            plt.text(
                -0.5 * readout_tone / self._x_unit[x_unit],
                ymax / 2.0,
                "readout",
                horizontalalignment="center",
                verticalalignment="center",
                rotation=90,
                size=14,
            )
        # adjust bounds
        plt.xlim(xmin + 0.005 * abs(xmax - xmin), xmax - 0.006 * abs(xmax - xmin))
        plt.ylim(ymin, ymax + 0.025 * (ymax - ymin))
        plt.xlabel("time " + x_unit)
        return

    def _get_boundaries(self, sequences, readout_indices, x_unit):
        """
        Returns plot boundaries for a given x-unit.

        Args:
            sequences:       list of all sequences stored in the channel
            readout_indices: indices of the readout for each sequence in sequences
            x_unit:          see dictionary

        Returns:
            List of boundaries (xmin, xmax, ymin, ymax)
        """
        try:
            readout_tone = self._sample.rec_pulselength
        except AttributeError:
            readout_tone = self._sample.readout_tone_length

        xmin = max(readout_indices) / (self._x_unit[x_unit] * self._sample.clock)
        xmax = (
            -(
                max(
                    np.array([len(seq) for seq in sequences])
                    - np.array(readout_indices)
                )
                / self._sample.clock
                + readout_tone
            )
            / self._x_unit[x_unit]
        )
        ymin = np.amin(np.concatenate(sequences))
        ymax = np.amax(np.concatenate(sequences))
        bounds = [xmin, xmax, ymin, ymax]
        return bounds

    def _get_sequences(self, IQ_mixing=False):
        """
        Explicitly calculate the waveforms of all sequences at their input time.

        Args:
            IQ_mixing: if true all waveforms are complex arrays where real and imaginary part encode I and Q

        Output:
            List of waveform arrays.
            List of readout indices.
        """

        try:
            samplerate = self._sample.clock
        except:
            samplerate = None

        seq_list = []
        ro_inds = []
        for sequence, variables in zip(self._sequences, self._variables):
            if variables:
                for single_variables in dictify_variable_lists(variables):
                    seq, ro_ind = sequence(
                        IQ_mixing=IQ_mixing, samplerate=samplerate, **single_variables
                    )
                    seq_list.append(seq)
                    ro_inds.append(ro_ind)
            else:
                # No variables
                seq, ro_ind = sequence(IQ_mixing=IQ_mixing, samplerate=samplerate)
                seq_list.append(seq)
                ro_inds.append(ro_ind)

        if not seq_list:
            logging.warning("No sequence stored in channel " + self.name)
            return [np.zeros(1)], [0]

        if self._interleave:
            seqs_temp = []
            ro_inds_temp = []
            time_dim = _vars_len(self._variables[0])
            for i in range(time_dim):
                seqs_temp += seq_list[i::time_dim]
                ro_inds_temp += ro_inds[i::time_dim]
            seq_list = seqs_temp
            ro_inds = ro_inds_temp
        return seq_list, ro_inds


class VirtualAWG(object):
    """
    Class managing multiple manipulation channels for a timedomain experiment (virtual AWG).
    Each channel stores pulse sequences which can be loaded onto your device.
    The readout pulse of each sequences is used to align sequences of multiple channles.

    Important attributes:
        self.channel[nr]: channel object - sequences can be added and plotted
        add_sequences:    add sequence(s) object (see pulse_sequence.py), maintaining currently stored sequences
        set_sequences:    add sequence(s) object (see pulse_sequence.py), deleting previously stored sequences
        delete_sequence:  delete previously stored sequences
        set_interleave:   for a specified channel turn interleaving of sequences on/off

        get_sequence_dicts: returns a dictionary containing all important channel attributes
        plot: plot the pulse sequences of all channels
        load: load sequences of specified channel(s) on your physical device

    TODO:
        Write .load fct to directly load the pulse sequences on a device, i.e. VirtualAWG.load(device).
    """

    def __init__(self, sample, channels=1):
        """
        Inits VirtualAWG with sample and number of channels:
            sample:   sample object
            channels: number of channels of the virtual AWG
        """
        self._sample = sample
        self._num_chans = channels

        self.channels = [None]  # type: List[TdChannel]
        self.channels.extend(
            TdChannel(self._sample, "channel_%i" % (i + 1)) for i in range(channels)
        )

        # Dictionary for x-axis scaling
        self._x_unit = {"s": 1, "ms": 1e-3, "us": 1e-6, "ns": 1e-9}
        # Dictonary for channel colors
        self._chancols = {1: "C0", 2: "C1", 3: "C2", 4: "C3", 5: "C4", 6: "C5", 7: "C6"}

    @property
    def channel_count(self):
        """The number of existing channels."""
        return self._num_chans

    def add_sequence(self, sequence, channel=1, **variables):
        """
        Sequence is added to current sequences.
        If channel is None all sequence is added to all channels.

        Args:
            sequence:     sequence or list of sequences.
            channel:      channel number in which sequence will be loaded.
            **variables:  keyword arguments matching the variable names of time functions and each containing a list of values.

        """
        if channel > self._num_chans:
            print(
                "Channel number {:d} is larger than total number of channels {:d}".format(
                    channel, self._num_chans
                )
            )
            return False
        if channel is not None:
            self.channels[channel].add_sequence(sequence, **variables)
        elif channel is None:
            for channel in self.channels:
                channel.add_sequence(sequence, **variables)
        return True

    def set_sequence(self, sequence, channel=1, **variables):
        """
        Delete previously added sequences, replacing them with sequence.
        If channel is None set all sequences in all channels to sequence.

        Args:
            sequence:     sequence or list of sequences.
            channel:      channel number in which sequence will be loaded.
            **variables:  keyword arguments matching the variable names of time functions and each containing a list of values.

        """
        if channel is not None:
            self.channels[channel].set_sequence(sequence, **variables)
        elif channel is None:
            for channel in self.channels:
                channel.set_sequence(sequence, **variables)
        return True

    def delete_sequence(self, seq_nr, channel=1):
        """
        Delete sequence seq_nr in stated channel.
        If channel is None sequence seq_nr is deleted from all channels.

        Args:
            seq_nr:  number of the sequence to be deleted
            channel: channel number
        """
        if channel is not None:
            self.channels[channel].delete_sequence(seq_nr)
        else:
            for chan in self.channels[1:]:
                chan.delete_sequence(seq_nr)
        return True

    def get_sequence_dicts(self):
        """
        For each channel, returns a dictionary featuring all important channel attributes:
            sequence_i: list of all pulses of sequnces number i (counting from 0)
            time_i: time values for which sequence is called
            value of the interleave attribute
        """
        chan_dicts = []
        for chan in self.channels[1:]:
            chan_dicts.append(chan.get_sequence_dict())
        return chan_dicts

    def set_interleave(self, value=True, channel=1):
        """
        For each channel, sequences with an equal number of timesteps may be interleaved.
        In order for this to work the squences must have the same number of time steps.

        Args:
            value:   if true, the sequences in this channel are interleaved.
            channel: number of the channel where sequences should be interleaved.
                     If channel is None action applies to all channels.
        """
        if channel is not None:
            self.channels[channel].set_interleave(value)
        else:
            for chan in self.channels[1:]:
                chan.set_interleave(value)
        return True

    def plot(self, show_quadrature="", x_unit="ns"):
        """
        Plots the sequences stored in all channels.
        A slider provides the option to sweep through the sequences.
        Readout pulse is fixed at t = 0, where is t>0 before the readout tone.
        By default, the amplitude of each sequence is plotted.
        This may be changed by setting the show_quadrature option to either "I" or "Q".

        Args:
            show_quadrature: set to "I" or "Q" if you want to display either quadrature instead of the amplitude
            x_unit:          unit of the x-axis in the plot (options are "s", "ms", "us", "ns").
        """
        seqs = []
        ro_inds = []
        seq_max = 0
        xmin, xmax, ymin, ymax = 0, 0, 0, 0

        show_iq = show_quadrature in ["I", "Q"]
        for chan in self.channels[1:]:
            sequences, readout_indices = chan._get_sequences(IQ_mixing=show_iq)

            # If there is no complex information in there, we just present the real part:
            has_complex_values = any([any(np.iscomplex(seq)) for seq in sequences])
            if show_quadrature == "I" or not has_complex_values:
                sequences = [np.real(seq) for seq in sequences]
            elif show_quadrature == "Q":
                sequences = [np.imag(seq) for seq in sequences]

            seq_max = max(seq_max, len(readout_indices) - 1)
            seqs.append(sequences)
            ro_inds.append(readout_indices)

            bounds = chan._get_boundaries(sequences, readout_indices, x_unit)
            xmin = max(xmin, bounds[0])
            xmax = min(xmax, bounds[1])
            ymin = min(ymin, bounds[2])
            ymax = max(ymax, bounds[3])

        interact(
            lambda sequence: self._plot_sequences(
                sequence, seqs, ro_inds, x_unit, [xmin, xmax, ymin, ymax]
            ),
            sequence=widgets.IntSlider(
                value=0, min=0, max=seq_max, layout=Layout(width="11in", height="50px")
            ),
        )
        return True

    def _plot_sequences(self, seq_ind, seqs, ro_inds, x_unit, bounds):
        """
        The actual plotting of the sequences happens here.

        Args:
            seqs:            list of sequences to be plotted (i.e. one list of sequence for each channel)
            ro_inds:         indices of the readout
            x_unit:          x_unit for the time axis
            bounds:          boundaries of the plot (xmin, xmax, ymin, ymax)
            show_quadrature: set to "I" or "Q" if you want to display either quadrature instead of the amplitude
        """
        if not PLOT_ENABLE:
            raise ImportError("matplotlib not found.")

        try:
            readout_tone = self._sample.rec_pulselength
        except AttributeError:
            readout_tone = self._sample.readout_tone_length

        fig = plt.figure(figsize=(18, 6))
        xmin, xmax, ymin, ymax = bounds
        samplerate = self._sample.clock
        # plot sequence
        for i, chan in enumerate(self.channels[1:]):
            if len(ro_inds[i]) > seq_ind:
                chan._plot_sequence(
                    seqs[i][seq_ind],
                    ro_inds[i][seq_ind],
                    x_unit,
                    bounds,
                    col=self._chancols[i + 1],
                    plot_readout=False,
                )

            # plot readout
        plt.fill(
            [
                0,
                0,
                -readout_tone / self._x_unit[x_unit],
                -readout_tone / self._x_unit[x_unit],
            ],
            [0, ymax, ymax, 0],
            color="C7",
            alpha=0.3,
        )
        # add label for readout
        plt.text(
            -0.5 * readout_tone / self._x_unit[x_unit],
            ymax / 2.0,
            "readout",
            horizontalalignment="center",
            verticalalignment="center",
            rotation=90,
            size=14,
        )

        # adjust plot limits
        plt.xlim(xmin + 0.005 * abs(xmax - xmin), xmax - 0.006 * abs(xmax - xmin))
        plt.ylim(ymin, ymax + 0.025 * (ymax - ymin))
        plt.xlabel("time " + x_unit)
        return

    def _get_sequences(self, IQ_mixing=False):
        """
        Returns a list of waveforms for each channel, as well as a list of indices for the readout timing.

        Args:
            IQ_mixing: if true all waveforms are complex arrays where real and imaginary part encode I and Q

        Returns:
            list of lists (one for each channel) of numpy arrays
        """
        seqs = []
        ro_inds = []
        for chan in self.channels[1:]:
            sequences, readout_indices = chan._get_sequences(IQ_mixing)
            seqs.append(sequences)
            ro_inds.append(readout_indices)
        return seqs, ro_inds

    def _sync(self):
        """
        Synchronize sequences of all channels to the same readout time.

        Output:
            sequences:       list of synchronized sequences as arrays (one each channel)
            readout_indices: array of new readout indices
        """
        seqs, ro_inds = self._get_sequences(IQ_mixing=True)
        # find maximum readout indices
        lens = [len(ro_ind) for ro_ind in ro_inds]
        readout_indices = np.zeros(max(lens))
        for i in range(0, len(lens)):
            for j in range(lens[i]):
                readout_indices[j] = max(readout_indices[j], ro_inds[i][j])

        ind = 0
        sequences = []
        sequences.extend([] for seq in seqs)
        while any(seqs):
            for i, seq in enumerate(seqs):
                if seq:
                    s = seq.pop(0)
                    temp = np.zeros(int(max(0, readout_indices[ind] - ro_inds[i][ind])))
                    s = np.append(temp, s)
                else:
                    s = np.zeros(0)
                sequences[i].append(s)
            ind += 1
        return sequences, readout_indices

    def load(self, show_progress_bar=True):
        """
        Load the sequences stored in the channels of the virtual AWG to your physical device (awg, fpga).
        Currently only enabled for the tabor awg.
        """
        # Case discrimination:
        if self._sample.awg.get_name() == "tawg":
            sequences, readout_inds = self._sync()
            load_tawg.load_tabor(
                sequences,
                readout_inds,
                self._sample,
                show_progress_bar=show_progress_bar,
            )
        else:
            print("Unknown device type! Unable to load sequences.")
        return True



# Anritsu_VNA.py
# hacked by Hannes Rotzinger hannes.rotzinger@kit.edu, 2011
# derived from Anritsu_VNA.py (and whatever this is derived from)
# Pascal Macha <pascalmacha@googlemail.com>, 2010
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from qkit.core.instrument_base import Instrument
from qkit import visa
import types
import logging
from time import sleep
import numpy

class Agilent_PNAX(Instrument):
    '''
    This is the python driver for the Agilent VNA X Vector Network Analyzer

    Usage:
    Initialise with
    <name> = instruments.create('<name>', address='<GPIB address>', reset=<bool>)
    
    '''

    def __init__(self, name, address, channel_index = 1):
        '''
        Initializes 

        Input:
            name (string)    : name of the instrument
            address (string) : GPIB address
        '''
        
        logging.info(__name__ + ' : Initializing instrument')
        Instrument.__init__(self, name, tags=['physical'])

        self._address = address
        self._visainstrument = visa.instrument(self._address)
        self._zerospan = False
        self._freqpoints = 0
        self._ci = channel_index 
        self._start = 0
        self._stop = 0
        self._nop = 0

        # Implement parameters
        self.add_parameter('nop', type=int,
            flags=Instrument.FLAG_GETSET,
            minval=1, maxval=100000,
            tags=['sweep'])
            
        self.add_parameter('bandwidth', type=float,
            flags=Instrument.FLAG_GETSET,
            minval=0, maxval=1e9,
            units='Hz', tags=['sweep']) 

        self.add_parameter('averages', type=int,
            flags=Instrument.FLAG_GETSET,
            minval=1, maxval=1024, tags=['sweep'])                    

        self.add_parameter('Average', type=bool,
            flags=Instrument.FLAG_GETSET)   
                    
        self.add_parameter('centerfreq', type=float,
            flags=Instrument.FLAG_GETSET,
            minval=0, maxval=20e9,
            units='Hz', tags=['sweep'])
            
        self.add_parameter('startfreq', type=float,
            flags=Instrument.FLAG_GETSET,
            minval=0, maxval=20e9,
            units='Hz', tags=['sweep'])            
            
        self.add_parameter('stopfreq', type=float,
            flags=Instrument.FLAG_GETSET,
            minval=0, maxval=20e9,
            units='Hz', tags=['sweep'])                        
            
        self.add_parameter('span', type=float,
            flags=Instrument.FLAG_GETSET,
            minval=0, maxval=20e9,
            units='Hz', tags=['sweep'])        
            
        self.add_parameter('power', type=float,
            flags=Instrument.FLAG_GETSET,
            minval=-30, maxval=23,
            units='dBm', tags=['sweep'])

        self.add_parameter('zerospan', type=bool,
            flags=Instrument.FLAG_GETSET)
            
        self.add_parameter('channel_index', type=int,
            flags=Instrument.FLAG_GETSET)            
                    
        #Triggering Stuff
        self.add_parameter('trigger_source', type=str,
            flags=Instrument.FLAG_GETSET)
            
        self.add_parameter('sweeptime', type=float,
            flags=Instrument.FLAG_GET,
            minval=0, maxval=200,
            units='s', tags=['sweep'])
            
        self.add_parameter('sweeptime_averages', type=float,
            flags=Instrument.FLAG_GET,
            minval=0, maxval=1000,
            units='s', tags=['sweep'])
            
        self.add_parameter('edel', type=float,
            flags=Instrument.FLAG_GETSET,
            minval=-10., maxval=10.,
            units='s', tags=['sweep'])
            
        self.add_parameter('cw', type=bool,
            flags=Instrument.FLAG_GETSET)
            
        self.add_parameter('cwfreq', type=float,
            flags=Instrument.FLAG_GETSET,
            minval=0, maxval=20e9,
            units='Hz', tags=['sweep'])
            
        self.add_parameter('source_attenuation', type=int,
            flags=Instrument.FLAG_GETSET,channels=(1,2),
            minval=0, maxval=60,
            tags=['sweep'])
            
        self.add_parameter('source_power_start', type=float,
            flags=Instrument.FLAG_GETSET,channels=(1,2),
            minval=-2.9e1, maxval=3e1,
            tags=['sweep'])
            
        self.add_parameter('source_power_stop', type=float,
            flags=Instrument.FLAG_GETSET,channels=(1,2),
            minval=-2.9e1, maxval=3e1,
            tags=['sweep'])
            
        self.add_parameter('calibration_state', type=bool,
            flags=Instrument.FLAG_GETSET) 
        
        self.add_parameter('sweep_type', type=str,
            flags=Instrument.FLAG_GETSET,tags=['sweep']) 
            
        self.add_parameter('sweep_mode', type=str,
            flags=Instrument.FLAG_GETSET,tags=['sweep']) 
            
        self.add_parameter('power_nop', type=int,
            flags=Instrument.FLAG_GETSET,channels=(1,2),
            minval=0, maxval=60,
            tags=['sweep'])
        
        self.add_parameter('avg_type', type=str,
            flags=Instrument.FLAG_GETSET,tags=['sweep'])
        
        # sets the S21 setting in the PNA X
        self.define_S21()
        
        # Implement functions
        self.add_function('get_freqpoints')
        self.add_function('get_tracedata')
        self.add_function('init')
        self.add_function('set_S21')
        self.add_function('avg_clear')
        self.add_function('pre_measurement')
        self.add_function('start_measurement')
        self.add_function('ready')
        self.add_function('post_measurement')
        #self.add_function('avg_status')
        
        #self._oldspan = self.get_span()
        #self._oldnop = self.get_nop()
        #if self._oldspan==0.002:
        #  self.set_zerospan(True)
        
        self.get_all()
    
    def get_all(self):
        self.get_nop()
        self.get_power()
        self.get_centerfreq()
        self.get_startfreq()
        self.get_stopfreq()
        self.get_span()
        self.get_bandwidth()
        self.get_trigger_source()
        self.get_Average()
        self.get_averages()
        self.get_freqpoints()   
        self.get_channel_index()
        self.get_cw()
        self.get_source_attenuation1()
        self.get_source_attenuation2()
        self.get_source_power_start1()
        self.get_source_power_start2()
        self.get_source_power_stop1()
        self.get_source_power_stop2()
        self.get_calibration_state()
        self.get_sweep_type()
        self.get_power_nop1()
        self.get_power_nop2()
        self.get_avg_type()
        self.get_edel()
        #self.get_zerospan()
        
    ###
    #Communication with device
    ###	
    
    def init(self):
        if self._zerospan:
          self._visainstrument.write('INIT1;*wai')
        else:
          if self.get_average():
            for i in range(self.get_averages()):            
              self._visainstrument.write('INIT1;*wai')
          else:
              self._visainstrument.write('INIT1;*wai')
              
    def hold(self, status):
        if status:
            self._visainstrument.write('SENSe:SWEep:MODE HOLD')
        else:
            self._visainstrument.write('SENSe:SWEep:MODE CONT')

    def set_S21(self):
        '''
        calls the defined S21 setting
        '''
        self._visainstrument.write("DISP:WIND:TRAC:FEED 'my_ch1_S21'")
        self._visainstrument.write("CALC:PAR:SEL 'my_ch1_S21'")

    def define_S21(self):
        '''
        defines the S21 measurement in the PNA X
        '''
        self._visainstrument.write( "CALCulate:PARameter:EXT 'my_ch1_S21','S21'")
        
        
    def reset_windows(self):
        self._visainstrument.write('DISP:WIND Off')
        self._visainstrument.write('DISP:WIND On')
    
    def set_autoscale(self):
        self._visainstrument.write("DISP:WIND:TRAC:Y:AUTO")
        
    def set_continous(self,ON=True):
        if ON:
            self._visainstrument.write( "INITiate:CONTinuous ON")
        else:
            self._visainstrument.write( "INITiate:CONTinuous Off")
    
    def get_sweep(self):
        self._visainstrument.write( "ABORT; INITiate:IMMediate;*wai")
        
    def avg_clear(self):
        self._visainstrument.write(':SENS%i:AVER:CLE' %(self._ci))

    def avg_status(self):
        # this does not work the same way than the VNA:
        #return int(self._visainstrument.ask(':SENS%i:AVER:COUN?' %(self._ci))
        pass
        
    def get_avg_status(self):
        return self._visainstrument.ask('STAT:OPER:AVER1:COND?')
            
    def still_avg(self): 
        if int(self.get_avg_status()) == 0: return True
        else: return False 
          
    def get_tracedata(self, format = 'AmpPha'):
        '''
        Get the data of the current trace

        Input:
            format (string) : 'AmpPha': Amp (lin) and Phase, 'RealImag',

        Output:
            'AmpPha': Amplitude [V] and Phase [pi]
        '''
        self._visainstrument.write(':FORMAT REAL,32; FORMat:BORDer SWAP;')
        self._visainstrument.write('CALC:FORM POL')
        data = self._visainstrument.ask_for_values( "CALCulate:DATA? FDATA",fmt = visa.single)
        data_size = numpy.size(data)
        datareal = numpy.array(data[0:data_size:2])
        dataimag = numpy.array(data[1:data_size:2])
        
        #print datareal,dataimag,len(datareal),len(dataimag)
        if format.upper() == 'REALIMAG':
          if self._zerospan:
            return numpy.mean(datareal), numpy.mean(dataimag)
          else:
            return datareal, dataimag
        elif format.upper() == 'AMPPHA':
          if self._zerospan:
            datareal = numpy.mean(datareal)
            dataimag = numpy.mean(dataimag)
            dataamp = numpy.sqrt(datareal*datareal+dataimag*dataimag)
            datapha = numpy.arctan(dataimag/datareal)
            return dataamp, datapha
          else:
            dataamp = numpy.sqrt(datareal*datareal+dataimag*dataimag)
            datapha = numpy.arctan2(dataimag,datareal)
            return dataamp, datapha
        else:
          raise ValueError('get_tracedata(): Format must be AmpPha or RealImag') 

      
    def get_freqpoints(self, query = False):      
      #if query == True:        
        #self._freqpoints = numpy.array(self._visainstrument.ask_for_values('SENS%i:FREQ:DATA:SDAT?'%self._ci,format=1)) / 1e9
        #self._freqpoints = numpy.array(self._visainstrument.ask_for_values(':FORMAT REAL,32;*CLS;CALC1:DATA:STIM?;*OPC',format=1)) / 1e9
      self._freqpoints = numpy.linspace(self._start,self._stop,self._nop)
      return self._freqpoints

    ###
    # SET and GET functions
    ###
    
    def do_set_nop(self, nop):
        '''
        Set Number of Points (nop) for sweep

        Input:
            nop (int) : Number of Points

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting Number of Points to %s ' % (nop))
        if self._zerospan:
          print 'in zerospan mode, nop is 1'
        else:
          self._visainstrument.write(':SENS%i:SWE:POIN %i' %(self._ci,nop))
          self._nop = nop
          self.get_freqpoints() #Update List of frequency points  
        
    def do_get_nop(self):
        '''
        Get Number of Points (nop) for sweep

        Input:
            None
        Output:
            nop (int)
        '''
        logging.debug(__name__ + ' : getting Number of Points')
        if self._zerospan:
          return 1
        else:
            self._nop = int(self._visainstrument.ask(':SENS%i:SWE:POIN?' %(self._ci)))    
        return self._nop 
    
    def do_set_Average(self, status):
        '''
        Set status of Average

        Input:
            status (boolean)

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting Average to "%s"' % (status))
        self._visainstrument.write('SENS%i:AVER:STAT %d' % (self._ci,status))
    def do_get_Average(self):
        '''
        Get status of Average

        Input:
            None

        Output:
            Status of Averaging (True or False) (boolean)
        '''
        logging.debug(__name__ + ' : getting average status')
        return bool(int(self._visainstrument.ask('SENS%i:AVER:STAT?' %(self._ci))))
                    
    def do_set_averages(self, av):
        '''
        Set number of averages and simultaneously sets this to the number of sweeps per group. This is important for sweep_mode(GRO)

        Input:
            av (int) : Number of averages

        Output:
            None
        '''
        if self._zerospan == False:
            logging.debug(__name__ + ' : setting Number of averages to %i ' % (av))
            self._visainstrument.write(':SENS%i:AVER:COUN %i; :SENS%i:SWE:GRO:COUN %i' % (self._ci,av,self._ci,av))
        else:
            self._visainstrument.write('SWE:POIN %.1f' % (self._ci,av)) #for zerospan, one would have to check what to take as group count
            
    def do_get_averages(self):
        '''
        Get number of averages

        Input:
            None
        Output:
            number of averages
        '''
        logging.debug(__name__ + ' : getting Number of Averages')
        if self._zerospan:
          return int(self._visainstrument.ask('SWE%i:POIN?' % self._ci))
        else:
          return int(self._visainstrument.ask('SENS%i:AVER:COUN?' % self._ci))
                
    def do_set_power(self,pow):
        '''
        Set probe power

        Input:
            pow (float) : Power in dBm

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting power to %s dBm' % pow)
        self._visainstrument.write('SOUR%i:POW1:LEV:IMM:AMPL %.1f' % (self._ci,pow))    
    def do_get_power(self):
        '''
        Get probe power

        Input:
            None

        Output:
            pow (float) : Power in dBm
        '''
        logging.debug(__name__ + ' : getting power')
        return float(self._visainstrument.ask('SOUR%i:POW1:LEV:IMM:AMPL?' % (self._ci)))
                
    def do_set_centerfreq(self,cf):
        '''
        Set the center frequency

        Input:
            cf (float) :Center Frequency in Hz

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting center frequency to %s' % cf)
        self._visainstrument.write('SENS%i:FREQ:CENT %f' % (self._ci,cf))
        self.get_startfreq();
        self.get_stopfreq();
        self.get_span();
    def do_get_centerfreq(self):
        '''
        Get the center frequency

        Input:
            None

        Output:
            cf (float) :Center Frequency in Hz
        '''
        logging.debug(__name__ + ' : getting center frequency')
        return  float(self._visainstrument.ask('SENS%i:FREQ:CENT?'%(self._ci)))
        
    def do_set_span(self,span):
        '''
        Set Span

        Input:
            span (float) : Span in KHz

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting span to %s Hz' % span)
        self._visainstrument.write('SENS%i:FREQ:SPAN %i' % (self._ci,span))   
        self.get_startfreq();
        self.get_stopfreq();
        self.get_centerfreq();   
    def do_get_span(self):
        '''
        Get Span
        
        Input:
            None

        Output:
            span (float) : Span in Hz
        '''
        #logging.debug(__name__ + ' : getting center frequency')
        span = self._visainstrument.ask('SENS%i:FREQ:SPAN?' % (self._ci) ) #float( self.ask('SENS1:FREQ:SPAN?'))
        return span

    def do_set_startfreq(self,val):
        '''
        Set Start frequency

        Input:
            span (float) : Frequency in Hz

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting start freq to %s Hz' % val)
        self._visainstrument.write('SENS%i:FREQ:STAR %f' % (self._ci,val))   
        self._start = val
        self.get_centerfreq();
        self.get_stopfreq();
        self.get_span();
    def do_get_startfreq(self):
        '''
        Get Start frequency
        
        Input:
            None

        Output:
            span (float) : Start Frequency in Hz
        '''
        logging.debug(__name__ + ' : getting start frequency')
        self._start = float(self._visainstrument.ask('SENS%i:FREQ:STAR?' % (self._ci)))
        return  self._start

    def do_set_stopfreq(self,val):
        '''
        Set STop frequency

        Input:
            val (float) : Stop Frequency in Hz

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting stop freq to %s Hz' % val)
        self._visainstrument.write('SENS%i:FREQ:STOP %f' % (self._ci,val))  
        self._stop = val
        self.get_startfreq();
        self.get_centerfreq();
        self.get_span();
    def do_get_stopfreq(self):
        '''
        Get Stop frequency
        
        Input:
            None

        Output:
            val (float) : Start Frequency in Hz
        '''
        logging.debug(__name__ + ' : getting stop frequency')
        self._stop = float(self._visainstrument.ask('SENS%i:FREQ:STOP?' %(self._ci) ))
        return  self._stop
               
    def do_set_bandwidth(self,band):
        '''
        Set Bandwidth

        Input:
            band (float) : Bandwidth in Hz

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting bandwidth to %s Hz' % (band))
        self._visainstrument.write('SENS%i:BWID:RES %i' % (self._ci,band))
    def do_get_bandwidth(self):
        '''
        Get Bandwidth

        Input:
            None

        Output:
            band (float) : Bandwidth in Hz
        '''
        logging.debug(__name__ + ' : getting bandwidth')
        # getting value from instrument
        return  float(self._visainstrument.ask('SENS%i:BWID:RES?'%self._ci))                

    def do_set_zerospan(self,val):
        '''
        Zerospan is a virtual "zerospan" mode. In Zerospan physical span is set to
        the minimal possible value (2Hz) and "averages" number of points is set.

        Input:
            val (bool) : True or False

        Output:
            None
        '''
        #logging.debug(__name__ + ' : setting status to "%s"' % status)
        if val not in [True, False]:
            raise ValueError('set_zerospan(): can only set True or False')        
        if val:
          self._oldnop = self.get_nop()
          self._oldspan = self.get_span()
          if self.get_span() > 0.002:
            Warning('Setting ZVL span to 2Hz for zerospan mode')            
            self.set_span(0.002)
            
        av = self.get_averages()
        self._zerospan = val
        if val:
            self.set_Average(False)
            self.set_averages(av)
            if av<2:
              av = 2
        else: 
          self.set_Average(True)
          self.set_span(self._oldspan)
          self.set_nop(self._oldnop)
          self.get_averages()
        self.get_nop()
               
    def do_get_zerospan(self):
        '''
        Check weather the virtual zerospan mode is turned on

        Input:
            None

        Output:
            val (bool) : True or False
        '''
        return self._zerospan


    def do_set_trigger_source(self,source):
        '''
        Set Trigger Mode

        Input:
            source (string) : AUTO | MANual | EXTernal | REMote

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting trigger source to "%s"' % source)
        if source.upper() in ['AUTO', 'MAN', 'EXT', 'REM']:
            self._visainstrument.write('TRIG:SOUR %s' % source.upper())        
        else:
            raise ValueError('set_trigger_source(): must be AUTO | MANual | EXTernal | REMote')
    def do_get_trigger_source(self):
        '''
        Get Trigger Mode

        Input:
            None

        Output:
            source (string) : AUTO | MANual | EXTernal | REMote
        '''
        logging.debug(__name__ + ' : getting trigger source')
        return self._visainstrument.ask('TRIG:SOUR?')        
        

    def do_set_channel_index(self,val):
        '''
        Set the index of the channel to address.

        Input:
            val (int) : 1 .. number of active channels (max 16)

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting channel index to "%i"' % int)
        nop = self._visainstrument.read('DISP:COUN?')
        if val < nop:
            self._ci = val 
        else:
            raise ValueError('set_channel_index(): index must be < nop channels')
    def do_get_channel_index(self):
        '''
        Get active channel

        Input:
            None

        Output:
            channel_index (int) : 1-16
        '''
        logging.debug(__name__ + ' : getting channel index')
        return self._ci
        
    def do_get_sweeptime_averages(self):
        return self.get_sweeptime() * self.get_averages()

    def do_get_sweeptime(self):

        '''
        Get sweep time

        '''

        logging.debug(__name__ + ' : getting sweep time')
        
        return float(self.get_nop()) / self.get_bandwidth()
        
        
    def do_set_edel(self, val, unit = 's'):
        '''
        Sets the electrical delay for the selected measurement.

        Input:
            val (int) :-10.00 - 10.00 [s]

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting electrical delay to "%g"' % val)
        self._visainstrument.write('SENS:CORR:EXT:PORT%i %g' %(self._ci, val))        
                                    
    def do_get_edel(self):
        '''
        Gets the electrical delay for the selected measurement.

        Input:
            None

        Output:
            val (int): -10.00 - 10.00 [s]
        '''
        logging.debug(__name__ + ' : getting electrical delay')
        return float(self._visainstrument.ask('SENS:CORR:EXT:PORT%i?' %(self._ci)))
        
    def do_set_cwfreq(self, cf):
        ''' set cw frequency '''
        self._visainstrument.write('SENS%i:FREQ:CW %f' % (self._ci,cf))

    def do_get_cwfreq(self):
        ''' get cw frequency '''
        return float(self._visainstrument.ask('SENS%i:FREQ:CW?'%(self._ci)))
        
        
    def do_set_cw(self, val):
        '''
        Set instrument to CW (single frequency) mode and back
        '''
        if val:
            self._visainstrument.write(':SENS%i:SWE:TYPE CW'%(self._ci))
        else:
            self._visainstrument.write(':SENS%i:SWE:TYPE LIN'%(self._ci))
        self._cw = val
        
    def do_get_cw(self):
        '''
        retrieve CW mode status from device
        '''
        if self._visainstrument.ask(':SENS%i:SWE:TYPE?'%(self._ci)) == 'CW':
            stat = True
        else:
            stat = False
        self._cw = stat
        return self._cw
        
    def do_get_source_attenuation(self,channel):
        '''
        Get status of source attenuation on port 

        Input:
            None

        Output:
            Status of source attenuation (value in dB) 
        '''
        
        logging.debug(__name__ + ' : getting attenuation status')
        return float(self._visainstrument.ask('SOUR%i:POW%i:ATT?' %(self._ci,channel)))

    def do_set_source_attenuation(self, att,channel):
        '''
        Set value of source attenuation

        Input:
            att (int)   :   Value of source attenuation must
                            must be one of (0,10,20,30,40,50,60)

        Output:
            None
        '''
        if att in (0,10,20,30,40,50,60):
            logging.debug(__name__ + ' : setting attenuation on port %i to %idB ' % (channel,att))
            self._visainstrument.write('SOUR%i:POW%i:ATT %i' % (self._ci,channel,att))
        else:
            logging.debug(__name__ + ' : cannot set attenuation to %i ' % (att))

    def do_get_source_power_start(self,channel):
        '''
        Get starting value for power sweep 

        Input:
            None

        Output:
            Starting value for the power sweep (float value in dBm) 
        '''
        
        logging.debug(__name__ + ' : getting starting power')
        return float(self._visainstrument.ask('SOUR%i:POW%i:PORT:STAR?' %(self._ci,channel)))
                    
    def do_set_source_power_start(self, pow,channel):
        '''
        Set starting value for power sweep 

        Input:
            pow (float)   :   Starting power (-29.9 to 30)

        Output:
            None
        '''
        
        logging.debug(__name__ + ' : setting starting power on port %i to %.2f dBm ' % (channel,pow))
        self._visainstrument.write('SOUR%i:POW%i:PORT:STAR %.2f' % (self._ci,channel,pow))
        
    def do_get_source_power_stop(self,channel):
        '''
        Get stopping value for power sweep 

        Input:
            None

        Output:
            Stopping value for the power sweep (float value in dBm) 
        '''
        
        logging.debug(__name__ + ' : getting stopping power')
        return float(self._visainstrument.ask('SOUR%i:POW%i:PORT:STOP?' %(self._ci,channel)))
                    
    def do_set_source_power_stop(self, pow,channel):
        '''
        Set stopping value for power sweep 

        Input:
            pow (float)   :   Stopping power (-29.9 to 30)

        Output:
            None
        '''
        
        logging.debug(__name__ + ' : setting stopping power on port %i to %.2f dBm ' % (channel,pow))
        self._visainstrument.write('SOUR%i:POW%i:PORT%:STOP %.2f' % (self._ci,channel,pow))
        
      
    def do_set_calibration_state(self, status):
        '''
        Set status of Calibration

        Input:
            status (boolean) 

        Output:
            None
        '''
        logging.debug(__name__+ ' : setting calibration stat to %i'%(status))
        self._visainstrument.write('SENS%i:CORR:STAT %i' % (self._ci))
        
    def do_get_calibration_state(self):
        '''
        Get status of Calibration

        Input:
            None

        Output:
            Status of Calibration (bool)
        '''
        logging.debug(__name__ + ' : getting calibration state status')
        return bool(int(self._visainstrument.ask('SENS%i:CORR:STAT?' %(self._ci))))

    def do_get_sweep_type(self):
        '''
        Get the Sweep Type

        Input:
            None

        Output:
                LIN:    Frequency-based linear sweep
                LOG:    Frequency-based logarithmic sweep
                CW:     Frequency-based continuous wave sweep
                SEG:    Segment-based sweep with frequency-based segments
                POW:    Power-based sweep with either Power-based sweep with a 
                    CW frequency, or Power-based sweep with swept-frequency
                PHAS:
        '''
        logging.debug(__name__ + ' : getting sweep type')
        
        return str(self._visainstrument.ask('SENS%i:SWE:TYPE?' %(self._ci)))

    def do_set_sweep_type(self,swtype):
        '''
        Set the Sweep Type
        Input:
            swtype (string):    One of
                LIN:    Frequency-based linear sweep
                LOG:    Frequency-based logarithmic sweep
                CW:     Frequency-based continuous wave sweep
                SEG:    Segment-based sweep with frequency-based segments
                POW:    Power-based sweep with either Power-based sweep with a 
                    CW frequency, or Power-based sweep with swept-frequency
                PHAS:   

        Output:
            None            
        '''
        if swtype in ('LIN','LOG','CW','SEGM','POW','PHAS'):
            
            logging.debug(__name__ + ' : Setting sweep type to %s'%(swtype))
            return self._visainstrument.write('SENS%i:SWE:TYPE %s' %(self._ci,swtype))
            
        else:
            logging.debug(__name__ + ' : Illegal argument %s'%(swtype))
            

    def do_get_power_nop(self,channel):
        '''
        Get number of points for a power sweep 

        Input:
            None

        Output:
            Number of points for a power sweep (int) 
        '''
        
        logging.debug(__name__ + ' : getting power nop')
        return int(self._visainstrument.ask('SENS%i:SWE:POIN?' %(self._ci)))




    def do_set_power_nop(self, n,channel):
        '''
        Set number of points for a power sweep 

        Input:
            n (int)   :   Number of points of the sweep

        Output:
            None
        '''
        
        logging.debug(__name__ + ' : setting number of power sweep points on channel %i to %i ' % (channel,n))
        self._visainstrument.write('SENS%i:SWE:POIN %i' % (self._ci,n))
        
    def do_get_avg_type(self):
        '''
        Get the Averaging Type

        Input:
            None

        Output:
            Averaging Type (string). One of
            POIN:	Point by Point
            SWE:	Sweep by Sweep
            
        '''
        logging.debug(__name__ + ' : getting averaging type')
        
        return str(self._visainstrument.ask('SENS%i:AVER:MODE?' %(self._ci)))

    def do_set_avg_type(self,avgtype):
        '''
        Set the Averaging Type
        Input:
            avgtype (string). One of:
                POIN:	Point by Point
                SWE:	Sweep by Sweep

        Output:
            None            
        '''
        if avgtype in ('POIN','SWE'):
            
            logging.debug(__name__ + ' : Setting avg type to %s'%(avgtype))
            return self._visainstrument.write('SENS%i:AVER:MODE %s' %(self._ci,avgtype))
            
        else:
            logging.debug(__name__ + ' : Illegal argument %s'%(swtype))
    def do_set_sweep_mode(self,mode):
        '''
        mode can be:
        CONT: Continuous sweep
        SING: Start a single scan (no averaging) and then hold.
        GRO:  Scan group of traces (ie number of averages) and then hold
        HOLD: finish current trace, then hold.
        '''
        if mode in ('CONT','SING','GRO','HOLD'):
            
            logging.debug(__name__ + ' : Setting sweep mode to %s'%(mode))
            return self._visainstrument.write('SENS%i:SWE:MODE %s' %(self._ci,mode))
            
        else:
            logging.warning(__name__ + ' : Illegal argument for sweep mode %s'%(mode)) 
     
    def do_get_sweep_mode(self):
        logging.debug(__name__ + ' : getting sweep type')
        return str(self._visainstrument.ask('SENS%i:SWE:MODE?' %(self._ci)))

    def pre_measurement(self):
        '''
        Set everything needed for the measurement
        '''
        self._visainstrument.write(':SENS%i:SWE:GRO:COUN %i' % (self._ci,self.get_averages())) #set the number of averages per grouped sweep
        
    def post_measurement(self):
        '''
        After a measurement, the VNA is in hold mode, and it can be difficult to start a measurement again from front panel.
        This function brings the VNA back to normal measuring operation.
        '''
        self.set_sweep_mode('CONT')
        
    def start_measurement(self):
        '''
        This function is called at the beginning of each single measurement in the spectroscopy script.
        Here, it resets the averaging and starts a GROUP sweep, which takes the specified number of averages.
        '''
        self.avg_clear()
        self.set_sweep_mode('GRO')
    
    def ready(self):
        '''
        This is a proxy function, returning True when the VNA has finished the required number of averages.
        '''
        return self.get_sweep_mode() == "HOLD"
        
    
        
    def read(self):
        return self._visainstrument.read()
    def write(self,msg):
        return self._visainstrument.write(msg)    
    def ask(self,msg):
        return self._visainstrument.ask(msg)



# Caen Fast-PS class, to perform the communication between the Wrapper and the device
# Marco Pfirrmann, marco.pfirrmann@kit.edu 2017
#
# based on the work on the Keithley_2636A driver by
# Pieter de Groot <pieterdegroot@gmail.com>, 2008
# Martijn Schaafsma <qtlab@mcschaafsma.nl>, 2008
# Hannes Rotzinger, hannes.rotzinger@kit.edu 2010
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from qkit.core.instrument_base import Instrument
import types
import socket
import logging
import time
import numpy

class Caen_FAST_PS(Instrument):
    '''
    This is the driver for the Caen Fast-PS Power Supply

    Usage:
    Initialize with
    <name> = instruments.create('<name>', 'Caen_fast_ps', address='<GPIB address>', reset=<bool>')
    '''

    def __init__(self, name, address, port = 10001):
        '''
        Initializes the Caen_fast_ps, and communicates with the wrapper.

        Input:
          name (string)    : name of the instrument
          address (string) : GPIB address
          reset (bool)     : resets to default values, default=False
        '''
        logging.info(__name__ + ' : Initializing instrument Caen_fast_ps')
        Instrument.__init__(self, name, tags=['physical'])

        self._host = address
        self._port = port
        self._soc = socket.socket(socket.AF_INET,socket.SOCK_STREAM)
        self._soc.connect((self._host, self._port))
        
        self._ak_str = '#AK'

        
        self.add_parameter('voltage',
            flags=Instrument.FLAG_GET, units='V', minval=-20, maxval=20, type=float)
            
        self.add_parameter('setvoltage',
            flags=Instrument.FLAG_GETSET, units='V', minval=-20, maxval=20, type=float)
 
        self.add_parameter('current',
            flags=Instrument.FLAG_GET, units='A', minval=-10, maxval=10, type=float)
            
        self.add_parameter('setcurrent',
            flags=Instrument.FLAG_GETSET, units='A', minval=-10, maxval=10, type=float)

        self.add_parameter('status',
            flags=Instrument.FLAG_GET, type=bool)

        self.add_parameter('output_mode',
            flags=Instrument.FLAG_GETSET, type=str)

        self.add_parameter('current_ramp_rate',
            flags=Instrument.FLAG_GETSET, units='A/s', minval=2e-4, maxval=10, type=float)

        self.add_parameter('voltage_ramp_rate',
            flags=Instrument.FLAG_GETSET, units='A/s', minval=-10, maxval=10, type=float)

        self.add_parameter('floating_mode', flags=Instrument.FLAG_GETSET, type=str)
        
        self.add_function('ramp_current')
        self.add_function('ramp_voltage')
        self.add_function('on')
        self.add_function('off')

    def get_all(self):
        '''
        Reads all implemented parameters from the instrument,
        and updates the wrapper.

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : get all')
        self.get('voltage')
        self.get('current')
        self.get('status')
        self.get('output_mode')
        self.get('setcurrent')
        self.get('setvoltage')
        self.get('current_ramp_rate')
        self.get('voltage_ramp_rate')
 
    def do_get_voltage(self):
        '''
        Reads the voltage signal from the instrument

        Input:
            None

        Output:
            volt (float) : Voltage in Volts
        '''
        logging.debug(__name__ + ' : get voltage')
        recv = self._ask('MRV:?')
        return float(recv.split(':')[1])

    def do_get_setvoltage(self):
        '''
        Reads the setvoltage signal from the instrument

        Input:
            None

        Output:
            setvolt (float) : Voltage in Volts
        '''
        logging.debug(__name__ + ' : get setvoltage')
        recv = self._ask('MWV:?')
        return float(recv.split(':')[1])

    def do_set_setvoltage(self, setvoltage):
        '''
        Reads the setvoltage signal from the instrument

        Input:
            setvoltage (float) : setvoltage in volts

        Output:
            None
        '''
        logging.debug(__name__ + ' : set setvoltage to %s' % setvoltage)
        recv = self._ask('MWV:%s' %setvoltage)
        if recv == self._ak_str: return True
        else: return 'ERROR: ' + error_msg[recv.split(':')[1]]
        
    def do_get_current(self):
        '''
        Reads the current signal from the instrument

        Input:
            None

        Output:
            current (float) : current in amps
        '''
        logging.debug(__name__ + ' : get current')
        recv = self._ask('MRI:?')
        return float(recv.split(':')[1])
        

    def do_get_setcurrent(self):
        '''
        Reads the setcurrent signal from the instrument

        Input:
            None

        Output:
            setcurrent (float) : setcurrent in amps
        '''
        logging.debug(__name__ + ' : get setcurrent')
        recv = self._ask('MWI:?')
        return float(recv.split(':')[1])
    
    def do_set_setcurrent(self, setcurrent):
        '''
        Reads the setcurrent signal from the instrument

        Input:
            setcurrent (float) : setcurrent in amps

        Output:
            None
        '''
        logging.debug(__name__ + ' : set setcurrent to %s' % setcurrent)
        recv = self._ask('MWI:%s' %setcurrent)
        if recv == self._ak_str:
            current = self.get_current()  # stores the actual current value in instrument.parameters
            return True
        else:
            return 'ERROR: ' + error_msg[recv.split(':')[1]]

    def do_get_status(self):
        '''
        Reads the output status from the instrument

        Input:
            None

        Output:
            status (int) : Usually True/False for On/Off, other integers specify certain errors. Conversion not (yet?) implemented.
        '''
        logging.debug(__name__ + ' : get status')
        recv = self._ask('MST:?')
        ret = int(recv.split(':')[1])
        return ret

    def do_get_output_mode(self):
        '''
        Reads the output mode from the instrument

        Input:
            None

        Output:
            mode (str) : 'V' oder 'I'
        '''
        logging.debug(__name__ + ' : get mode')
        recv = self._ask('LOOP:?')
        ret = recv.split(':')[1]
        return ret

    def do_set_output_mode(self, mode):
        '''
        Sets the output mode from the instrument
        ## There seems to be an error somewhere. The mode is set correctly but the return response does not work propperly

        Input:
            mode (str) : 'V' oder 'I'

        Output:
            None
        '''
        logging.debug(__name__ + ' : set mode')
        self.off()
        recv = self._ask('LOOP:%s' %mode)
        self.on()
        if recv == self._ak_str: return True
        else: return 'ERROR: ' + error_msg[recv.split(':')[1]]
        
    def do_get_current_ramp_rate(self):
        '''
        Reads the setcurrent signal from the instrument

        Input:
            None

        Output:
            rate (float) : current_ramp_rate in A/s
        '''
        logging.debug(__name__ + ' : get current ramp rate')
        recv = self._ask('MSRI:?')
        ret = recv.split(':')[1]
        return ret
    
    def do_set_current_ramp_rate(self, rate):
        '''
        Reads the setcurrent signal from the instrument

        Input:
            rate (float) : current_ramp_rate in A/s

        Output:
            None
        '''
        logging.debug(__name__ + ' : set current ramp rate to %s' %rate)
        recv = self._ask('MSRI:%s' %rate)
        if recv == self._ak_str: return True
        else: return 'ERROR: ' + error_msg[recv.split(':')[1]]

    def do_get_voltage_ramp_rate(self):
        '''
        Reads the setcurrent signal from the instrument

        Input:
            None

        Output:
            rate (float) : voltage_ramp_rate in V/s
        '''
        logging.debug(__name__ + ' : get voltage ramp rate')
        recv = self._ask('MSRV:?')
        ret = recv.split(':')[1]
        return ret
    
    def do_set_voltage_ramp_rate(self, rate):
        '''
        Reads the setcurrent signal from the instrument

        Input:
            rate (float) : voltage_ramp_rate in V/s

        Output:
            None
        '''
        logging.debug(__name__ + ' : set voltage ramp rate to %s' %rate)
        recv = self._ask('MSRV:%s' %rate)
        if recv == self._ak_str: return True
        else: return 'ERROR: ' + error_msg[recv.split(':')[1]]
        
    def do_set_floating_mode(self, mode):
        """
        Enables(Deables) the floating mode. However, earth fuse must be removed (inserted.)
        Input:
            mode (str) : 'N' = non floating, 'F' = floating
        Output:
            None
        """
        logging.debug(__name__ + ' : set floating mode to {}'.format(mode))
        if mode == 'N':
            print("WARNING: Earth fuse has to be inserted")
        elif mode == 'F':
            print("WARNING: Earth fuse has to be removed")
        else:
            print('wrong input')
            return
        recv = self._ask('SETFLOAT:{}'.format(mode))
        if recv == self._ak_str:
            return True
        else:
            return 'ERROR: ' + error_msg[recv.split(':')[1]]

    def do_get_floating_mode(self):
        """
        returns the floating mode
        Input: None
        Output (STR): F = floating, N = non-floating
        """
        logging.debug(__name__ + ' : get floating mode')
        recv = self._ask('SETFLOAT:?')
        ret = recv.split(':')[1]
        return ret

    # shortcuts
    def off(self):
        '''
        Sets status to False

        Input:
            None

        Output:
            None
        '''
        recv = self._ask('MOFF')
        if recv == self._ak_str: return True
        else: return 'ERROR: ' + error_msg[recv.split(':')[1]]

    def on(self):
        '''
        Sets status to True

        Input:
            None

        Output:
            None
        '''
        recv = self._ask('MON')
        if recv == self._ak_str: return True
        else: return 'ERROR: ' + error_msg[recv.split(':')[1]]
    
    def ramp_finished(self):
        time.sleep(0.1)
        while int(self._ask('MST:?').split(':')[1], 16) & 2**12 == 2**12: #MP: gets 32bit status array; 12th from last bit indicates ramping process
            time.sleep(0.1)
        return True
    
    def ramp_current(self, current, ramp_rate = False, wait = True):
        '''
        Ramps current with given rate
        
        Input:
            current (float) : destination value in A
            ramp_rate (float) : ramp rate in A/s (optional)
            wait (bool) : wait until set current value is reached before returning its value (recommended)
        
        Output:
            None
        '''
        if ramp_rate: self.do_set_current_ramp_rate(ramp_rate)
        logging.debug(__name__ + ' : ramp current to %s' %current)
        recv = self._ask('MWIR:%s' %current)
        if recv == self._ak_str: 
            if wait:
                time.sleep(0.1)
                while not self.ramp_finished():
                    time.sleep(0.1)
                time.sleep(0.1)
            current = self.get_current()  # stores the actual current value in instrument.parameters
            return True
        else:
            return 'ERROR: ' + error_msg[recv.split(':')[1]]
    
    def ramp_voltage(self, voltage, ramp_rate = False, wait = True):
        '''
        Ramps voltage with given rate
        
        Input:
            voltage (float) : destination value in V
            ramp_rate (float) : ramp rate in V/s (optional)
            wait (bool) : wait until set current value is reached before returning its value (recommended)
        
        Output:
            None
        '''
        if ramp_rate: self.do_set_voltage_ramp_rate(ramp_rate)
        logging.debug(__name__ + ' : ramp voltage to %s' %voltage)
        recv = self._ask('MWVR:%s' %voltage)
        if recv == self._ak_str:
            if not wait:
                return True
            else:
                time.sleep(0.1)
                while not self.ramp_finished():
                    time.sleep(0.1)
                time.sleep(0.1)
                return str(self.do_get_voltage())
       
    def _ask(self, cmd):
        cmd = (cmd+'\r').encode('utf-8')
        self._soc.sendall(cmd)
        return self._soc.recv(1024).strip().decode('utf8')
       
    def set_coil(self, coil):
        '''
        Adapts the PID values to the specific coils in our lab
        
        Input:
            coil (int) : {0:default, 1:solenoid_red, 2:yoke_red, 3:solenoid_janice}

        Output:
            None
        '''
        PID_dict = {0:[0.0001, 0.0001, 0], 1:[0.0001, 0.0001, 0], 2:[0.0001, 0.1, 0], 3:[0.0001, 0.0001, 0]}
        PID = PID_dict[coil]
        recv = self._ask('MWG:43:%f' %PID[0])
        if recv != self._ak_str: return 'ERROR: ' + error_msg[recv.split(':')[1]]
        recv = self._ask('MWG:44:%f' %PID[1])
        if recv != self._ak_str: return 'ERROR: ' + error_msg[recv.split(':')[1]]
        recv = self._ask('MWG:45:%f' %PID[2])
        if recv != self._ak_str: return 'ERROR: ' + error_msg[recv.split(':')[1]]
        recv = self._ask('MSAVE')
        if recv != self._ak_str: return 'ERROR: ' + error_msg[recv.split(':')[1]]   
        
        
error_msg = {'01':"Unknown command",
             '02':"Unknown Parameter",
             '03':"Index out of range",
             '04':"Not Enough Arguments",
             '05':"Privilege Level Requirement not met",
             '06':"Saving Error on device",
             '07':"Invalid password",
             '08':"Power supply in fault",
             '09':"Power supply already ON",
             '10':"Setpoint is out of model limits",
             '11':"Setpoint is out of software limits",
             '12':"Setpoint is not a number",
             '13':"Module is OFF",
             '14':"Slew Rate out of limits",
             '15':"Device is set in local mode",
             '16':"Module is not in waveform mode",
             '17':"Module is in waveform mode",
             '18':"Device is set in remote mode",
             '19':"Module is already in the selected loop mode",
             '20':"Module is not in the selected loop mode",
             '99':"Unknown error"}



# filename: Keithley_2000.py
# version 0.1 written by JB,JNV, HR@KIT 2017-
# QKIT driver for a Keithley Multimeter 2000 

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


import qkit
from qkit.core.instrument_base import Instrument

import logging
import numpy
import time,sys
import atexit
import serial

class Keithley_2000(Instrument):
    '''
    This is the driver for the Keithley 2000 Source Meter

    Usage:
    Initialize with
    <name> = qkit.instruments.create('<name>', 'Keithley_2000', address='<GBIP address>', reset=<bool>)
    '''

    def __init__(self, name, address, reset=False):
        '''
        Initializes the Keithley, and communicates with the wrapper.
        
        Input:
            name (string)    : name of the instrument
            address (string) : GPIB address
            reset (bool)     : resets to default values, default=False
        '''
        # Start VISA communication
        logging.info(__name__ + ': Initializing instrument Keithley 2000')
        Instrument.__init__(self, name, tags=['physical'])
        #self._address = address
        
        self._rvol = 1
        self.rvc_mode   = False
        self.four_wire  = False
        self.setup(address)

    def setup(self, device="/dev/ttyUSB7"):
        baudrate = 9600
        timeout = 0.1
        
        self.ser = self._std_open(device,baudrate,timeout)
        atexit.register(self.ser.close)
        
    def _std_open(self,device,baudrate,timeout):
        # open serial port, 9600, 8,N,1, timeout 0.1
        return serial.Serial(device, baudrate, timeout=timeout)
        
    def remote_cmd(self, cmd):
        cmd += "\r"

        # clear queue first, old data,etc
        rem_char = self.ser.inWaiting()
        if rem_char:
            self.ser.read(rem_char)
        
        # send command
        self.ser.write(str.encode(cmd))
        # wait until data is processed
        time.sleep(1)
        # read back
        rem_char = self.ser.inWaiting()
        
        retval = self.ser.read(rem_char)
        #print(retval)
        return retval #str(retval)#.strip('\r')
    
    def get_current_dc(self):   #Return DC Current in auto-range
        value = self.remote_cmd(":MEAS:CURR:DC?")
        try:
            return float(value)
        except Exception as m:
            print(m)
            return 

    def get_resistance(self):
        if self.rvc_mode:
            return self._rvol/self.get_data()
        else:
            if self.four_wire:
                return self.get_resistance_4W()
            else:
                return self.get_resistance_2W()
                
    def get_data(self):
        try:
            ret = self.remote_cmd(":DATA?")
            return float(ret)
        except ValueError as e:
            print(e)
            print(ret)
            return numpy.NaN        

    def get_resistance_2W(self):
        try:
            ret = self.remote_cmd(":MEAS:RES?")
            return float(ret)
        except ValueError as e:
            print(e)
            print(ret)
            return numpy.NaN

    def get_resistance_4W(self):
        try:
            ret = self.remote_cmd(":MEAS:FRES?")
            return float(ret)
        except ValueError as e:
            print(e)
            print(ret)
            return numpy.NaN

    def set_measure_4W(self,four_wire):
        ''' Sets 2 or 4 wire measurement mode '''
        self.four_wire = four_wire
        
    def set_resistance_via_current(self, status):
        self.rvc_mode = status
        self.ser.write(str.encode(":FUNC 'CURR:DC' \r"))
        self.ser.write(str.encode(":CURR:DC:RANG:AUTO OFF \r"))
        self.ser.write(str.encode(":CURR:DC:RANG 1e-3 \r"))

    def set_reverence_voltage(self, voltage):
        self._rvol = voltage
        
    def set_current_range(self, range_value):
        if 1e-3 <= range_value <= 1:
            self.ser.write(str.encode(":CURR:DC:RANG %.04f \r"%(range_value)))
        else:
            print("Range must be decimal power of 10 between 1A and 10mA")

if __name__ == "__main__":
    KEITH = Keithley_2000(name = "Keithley_2000", address="COM6")
    print("DC current: {:.4g}A".format(KEITH.get_current_dc()))



# -*- coding: utf-8 -*-
"""
Library to ease the use of the file format hdf5 for datastorage
It can be used with or without the qtlab environment.

@author: hannes.rotzinger@kit.edu 2015
@version: 0.1
"""
import logging
import os
import time

import qkit
from qkit.storage.hdf_file import H5_file
from qkit.storage.hdf_dataset import hdf_dataset
from qkit.storage.hdf_constants import ds_types
from qkit.storage.hdf_view import dataset_view
from qkit.storage.hdf_DateTimeGenerator import DateTimeGenerator
import warnings
# this module shouldt be imported, please... don't use it...
warnings.warn(
   "This module is deprecated and will be removed in the near future. It was replaced by the module 'qkit.storage.store'.",
    FutureWarning
)


class Data(object):
    "this is a basic hdf5 class adopted to our needs"
    # a types
    def __init__(self, *args, **kwargs):
        """
        Creates an empty data set including the file, for which the currently
        set file name generator is used.

        kwargs:
            name (string):  default is 'data', will result in a file at datadir/uuid_name.h5
            path (string): to open an existing file or specify the exact path. If None, the file will be created.
        """

        self._name = kwargs.pop('name', 'data')

        #if path was omitted, a new filepath will be created
        self._filepath = kwargs.pop('path',None)
        
        if self._filepath is None:
            self.generate_file_name()
        else:
            self._filepath = os.path.abspath(self._filepath)
            self._folder,self._filename = os.path.split(self._filepath)
        "setup the  file"
        self.hf = H5_file(self._filepath,mode='r+')
        self.hf.flush()


    def generate_file_name(self):
        dtg = DateTimeGenerator()
        self.__dict__.update(dtg.new_filename(self._name))
        '''
        this sets:
            _unix_timestamp
            _localtime    
            _timestamp
            _timemark
            _datemark
            _uuid
            _filename
            _folder
            _relpath
            _filepath        
        '''
        
        if self._folder and not os.path.isdir(self._folder):
            os.makedirs(self._folder)

    def __getitem__(self, name):
        return self.hf[name]

    def __setitem__(self, name, val):
        self.hf[name] = val

    def __repr__(self):
        ret = "HDF5Data '%s', filename '%s'" % (self._name, self._filename)
        return ret

    def get_filepath(self):
        return self._filepath

    def get_folder(self):
        return self._folder

    def add_comment(self,comment, folder = "data" ):
        if folder == "data":
            #existing_comment = self.hf.dgrp.attrs.get('comment',None)
            #if existing_comment:
            #    comment = existing_comment+\n+comment
            self.hf.dgrp.attrs.create('comment',comment)
        if folder == "analysis":
            #existing_comment = self.hf.agrp.attrs.get('comment',None)
            #if existing_comment:
            #    comment = existing_comment+\n+comment
            self.hf.agrp.attrs.create("comment",comment)
    def add_textlist(self,name,comment = "" ,folder="data", **meta):
        ds =  hdf_dataset(self.hf, name, comment = comment, folder=folder, ds_type = ds_types['txt'], dim=1, **meta)
        return ds
        
    def add_coordinate(self,  name, unit = "", comment = "",folder="data",**meta):
        ds =  hdf_dataset(self.hf, name,unit=unit, ds_type = ds_types['coordinate'],
                          comment= comment, folder=folder, dtype='float64', **meta)
        return ds

    def add_value_vector(self, name, x = None, unit = "", comment = "",folder="data",**meta):
        ds =  hdf_dataset(self.hf, name, x=x, unit=unit, ds_type = ds_types['vector'],
                          comment=comment, folder=folder,**meta)
        return ds

    def add_value_matrix(self, name, x = None , y = None, unit = "", comment = "",folder="data",**meta):
        ds =  hdf_dataset(self.hf, name, x=x, y=y, unit=unit, ds_type = ds_types['matrix'],
                          comment=comment, folder=folder,**meta)
        return ds

    def add_value_box(self, name, x = None , y = None, z = None, unit = "", comment = "",folder="data",**meta):
        ds =  hdf_dataset(self.hf,name, x=x, y=y, z=z, unit=unit, ds_type = ds_types['box'],
                          comment=comment, folder=folder,**meta)
        return ds

    def add_view(self,name,x = None, y = None, error = None, filter  = None, view_params = {}):
        """
        a view is a way to display plot x-y data.
        x, y with the corresponding error are the datasets to display, e.g.
        x = h5["/entry/data0/temperature"]
        y = h5["/entry/analysis0/frequency_fit"]
        error = h5["/entry/analysis0/frequency_fit_error"]
        filter is a string of reguar python code, which
        accesses the x,y dataset returns arrays of (x,y)
        (Fixme: not jet implemented)
        """
        ds =  dataset_view(self.hf,name, x=x, y=y, error=error, filter = filter, 
                           ds_type = ds_types['view'],view_params = view_params)
        return ds

    def get_dataset(self,ds_url):
        return hdf_dataset(self.hf,ds_url = ds_url)

    def save_finished():
        pass

    def flush(self):
        self.hf.flush()

    def close_file(self):
        self.hf.close_file()
    def close(self):
        self.hf.close_file()



# Quantum machines coherence library
# started by M. Spiecker, 06.2020
#
#
#

import inspect
import time
import threading

import qkit
from qkit.storage import store as hdf
from qkit.measure.measurement_class import Measurement
import qkit.measure.write_additional_files as waf
from qkit.gui.plot import plot as qviewkit


class QmQkitWrapper:

    def __init__(self):

        self.exp_name = None
        self.dirname = ""  # TODO new name
        self.comment = "test"
        self.sourcecode = ''
        self.config = ''
        self.sample = "tests"
        self.pulseinfo = "NoPulseDataSaved"
        self.scaleoffset = None

        self._measurement_object = Measurement()
        self._measurement_object.measurement_type = 'TimeDomain'




        # data
        self.coords = {}
        self.values = {}


    # Decorator to start qkit
    def measure(func):
        def wrapper(self, *args, **kwargs):

            save = False
            if 'save' in kwargs:
                save = kwargs['save']
                kwargs.pop('save', None)

            if save:
                qkit.flow.start()

            output = func(self, *args, **kwargs)

            # Save function arguments and qm program
            astring = ""
            for value in args:
                astring += "{}, ".format(value)

            kstring = ""
            for key, value in kwargs.items():
                kstring += "{} = {}, ".format(key, value)

            self.sourcecode += astring + kstring + "\n\n\n" + inspect.getsource(func)

            if save:
                if self.dirname:
                    self._file_name = 'QM_experiment_' + self.exp_name + '_' + self.dirname
                else:
                    self._file_name = 'QM_experiment_' + self.exp_name
                self._file_name = self._file_name.replace(' ', '').replace(',', '_')

                self._prepare_measurement_file()
                self.store_data()

                qkit.flow.end()

                self.close_files()

                #print('Measurement complete: {:s}'.format(self._data_file.get_filepath()))

            else:
                #print('Measurement complete')
                pass

            return output
        return wrapper

    def _prepare_measurement_file(self):
        '''
        creates the output .h5-file with distinct dataset structures for each measurement type.
        at this point all measurement parameters are known and put in the output file
        '''

        self._data_file = hdf.Data(name=self._file_name, mode='a')
        self._measurement_object.uuid = self._data_file._uuid
        self._measurement_object.hdf_relpath = self._data_file._relpath
        self._measurement_object.instruments = qkit.instruments.get_instrument_names()

        #self._measurement_object.sample = "testsample"
        #self._measurement_object.measurement_func = self.pulseinfo

        self._measurement_object.save()

        self._mo = self._data_file.add_textlist('measurement')
        self._mo.append(self._measurement_object.get_JSON())

        # instrument settings and logfile
        self._settings = self._data_file.add_textlist('settings')
        settings = waf.get_instrument_settings(self._data_file.get_filepath())
        self._settings.append(settings)

        self._log_file = waf.open_log_file(self._data_file.get_filepath())



    def close_files(self):

        # save plots and close files
        # t = threading.Thread(target=qviewkit.save_plots, args=[self._data_file.get_filepath()])
        # t.start()

        self._data_file.flush()
        self._data_file.close_file()
        waf.close_log_file(self._log_file)

        # TODO open qkit
        # if self.qviewkit_singleInstance and self.open_qviewkit and self._qvk_process:
        #    self._qvk_process.terminate()  # terminate an old qviewkit instance


    def store_data(self):
        """

        """
        coord_dic = {}

        for key in self.coords:
            coord_vec = self.coords[key][0]
            unit = self.coords[key][1]

            coords_file = self._data_file.add_coordinate(key, unit=unit)
            coords_file.add(coord_vec)
            coord_dic[key] = coords_file

        for key in self.values:
            values = self.values[key][0]
            coord_key_list = self.values[key][1]
            unit = self.values[key][2]

            if len(coord_key_list) == 1:
                value_file = self._data_file.add_value_vector(key, x=coord_dic[coord_key_list[0]], unit=unit, scaleoffset = self.scaleoffset)
                value_file.append(values)
            elif len(coord_key_list) == 2:
                value_file = self._data_file.add_value_matrix(key, x=coord_dic[coord_key_list[0]],
                                                              y=coord_dic[coord_key_list[1]], unit=unit, scaleoffset = self.scaleoffset)
                # file initialization - workaround
                value_file.append(values[0, :])
            elif len(coord_key_list) == 3:
                value_file = self._data_file.add_value_box(key, x=coord_dic[coord_key_list[0]],
                                                           y=coord_dic[coord_key_list[1]],
                                                           z=coord_dic[coord_key_list[2]], unit=unit, scaleoffset = self.scaleoffset)
                # file initialization - workaround
                value_file.append(values[0, 0, :])

            value_file.ds.resize(values.shape)
            value_file.ds[:] = values

        # source code
        sourcecode_file = self._data_file.add_textlist('sourcecode')
        sourcecode_file.append(self.sourcecode)

        self.sourcecode = ""

        config_file = self._data_file.add_textlist('config')
        config_file.append(self.qm_config.config)

        scale_offset_file = self._data_file.add_textlist('scale_offset')
        scale_offset_file.append(self.scaleoffset)

        # comment
        if self.comment:
            self._data_file.add_comment(self.comment)








# wrapper for EdwardsActiveDigitalController by JB@KIT 03/2015
# pressure readout device @ UFO: G1, G2
# model: D39591000 (enhanced)

from qkit.core.instrument_base import Instrument
import socket

class EdwardsActiveDigitalController(Instrument):

    def __init__(self, name, host='pi-us74', port=9955):

        Instrument.__init__(self, name, tags=['physical'])
        
        self.add_parameter('condenser_pressure',
            type=float,
            flags=Instrument.FLAG_GET, units='mbar')
        self.add_parameter('still_pressure',
            type=float,
            flags=Instrument.FLAG_GET, units='mbar')
            
        self.HOST, self.PORT = host, port

    def do_get_condenser_pressure(self):
        try:
            # Create a socket (SOCK_STREAM means a TCP socket)
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            #connect
            sock.connect((self.HOST, self.PORT))
            sock.sendall('get_p_cond' + '\n')
            # Receive data from the server and shut down
            rec = sock.recv(1024)
            sock.close()
            return rec
        except Exception as detail:
            print "Error: ", detail
            return 0
            
    def do_get_still_pressure(self):
        try:
            # Create a socket (SOCK_STREAM means a TCP socket)
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            #connect
            sock.connect((self.HOST, self.PORT))
            sock.sendall('get_p_still' + '\n')
            # Receive data from the server and shut down
            rec = sock.recv(1024)
            sock.close()
            return rec
        except Exception as detail:
            print "Error: ", detail
            return 0
    
    def get_all(self):
        self.get('condenser_pressure')
        self.get('still_pressure')
 



# Quantum machines coherence library
# started by M. Spiecker, 06.2020
#
# TODO add text

# HOW TO USE IT:
# Since the quantum machines card is a very versatile machine,
# the setup and the experiments will be very different.
# Thus use this file as template and look up for your experiments.



import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt

from qm.qua import *
from qm import SimulationConfig
from qm import LoopbackInterface

from qm.QuantumMachinesManager import QuantumMachinesManager
from qkit.measure.timedomain.quantum_machine.qm_qkit_wrapper import QmQkitWrapper

bit_shift = 12

class QuantumJumpMonitoring(QmQkitWrapper):

    def __init__(self, qm_config):

        QmQkitWrapper.__init__(self)

        self.qm_config = qm_config

        # global experiment settings
        self.wait_before = 250  # times 4 for ns

        self.program = None

        # Output
        self.I = None
        self.Q = None
        self.state = None
        self.t = None

        self.data = None

        # plots
        self.fig = None

    def set_figure(self, fig):
        self.fig = fig

    # Experiments

    @QmQkitWrapper.measure
    def quantum_jump_test(self, n_measure, ref_pos, dist, t_wait=25, plot=True):

        pos1x = ref_pos[0, 0].item()
        pos1y = ref_pos[0, 1].item()
        pos2x = ref_pos[1, 0].item()
        pos2y = ref_pos[1, 1].item()
        var1 = dist[0].item()
        var2 = dist[1].item()

        with program() as quantumJumps:
            N = declare(int)
            Is = declare(fixed)
            Qs = declare(fixed)
            Ir = declare(fixed)
            Qr = declare(fixed)
            I = declare(fixed)
            Q = declare(fixed)

            d1 = declare(fixed)
            d2 = declare(fixed)
            state = declare(bool)
            #state = declare(int)

            with for_(N, 0, N < n_measure, N + 1):

                # measurement block:
                measure("readout_sig", "signal", None, ("integW_Is", Is), ("integW_Qs", Qs))
                measure("readout_ref", "reference", None, ("integW_Ir", Ir), ("integW_Qr", Qr))

                wait(t_wait, "signal")

                # calculations:
                assign(I, (Is << bit_shift) * (Ir << bit_shift) + (Qs << bit_shift) * (Qr << bit_shift))
                assign(Q, (Is << bit_shift) * (Qr << bit_shift) - (Qs << bit_shift) * (Ir << bit_shift))

                assign(d1, (I - pos1x) * (I - pos1x) + (Q - pos1y) * (Q - pos1y))
                assign(d2, (I - pos2x) * (I - pos2x) + (Q - pos2y) * (Q - pos2y))

                assign(state, d1 < d2)
                # with if_(d1 < var1):
                #    assign(state, -1)
                # with else_():
                #    with if_(d2 < var2):
                #        assign(state, 1)
                #    with else_():
                #        assign(state, 0)

                # save to client:
                save(I, "I")
                save(Q, "Q")
                save(d1, "d1")
                save(d2, "d2")
                save(state, "state")

        self.program = quantumJumps
        job = self.qm_config.qm.execute(self.program, duration_limit=0, data_limit=0)
        job.wait_for_all_results()
        res = job.get_results()

        self.I = res.variable_results.I.values
        self.Q = res.variable_results.Q.values
        d1 = res.variable_results.d1.values
        d2 = res.variable_results.d2.values
        self.state = res.variable_results.state.values
        self.t = res.variable_results.state.ts_nsec

        # qkit data storage
        self.exp_name = "quantum_jumps_{}us".format(4 * t_wait / 1000).replace('.', 'p')
        self.coords = {"t": [self.t / 1000, "us"]}
        self.values = {"state": [state, ["t"], "sate"],
                       "I": [self.I, ["t"], "V"], "Q": [self.Q, ["t"], "V"]}

        if plot:
            self.fig.clear()
            grid = self.fig.add_gridspec(ncols=1, nrows=4, height_ratios=[1, 1, 1, 2])

            ax1 = self.fig.add_subplot(grid[0, 0])
            ax1.plot(np.arange(n_measure), self.I)
            ax1.plot(np.arange(n_measure), self.Q)

            ax2 = self.fig.add_subplot(grid[1, 0])
            ax2.plot(np.arange(n_measure), d1)
            ax2.plot(np.arange(n_measure), d2)

            ax3 = self.fig.add_subplot(grid[2, 0])
            ax3.plot(np.arange(n_measure), self.state)

            ax4 = self.fig.add_subplot(grid[3, 0])
            ax4.scatter(self.I, self.Q, 0.1)
            s = np.linspace(0, 2 * np.pi)
            ax4.plot(ref_pos[0, 0] + np.sqrt(var1) * np.cos(s), ref_pos[0, 1] + np.sqrt(var1) * np.sin(s), 'k')
            ax4.plot(ref_pos[1, 0] + np.sqrt(var2) * np.cos(s), ref_pos[1, 1] + np.sqrt(var2) * np.sin(s), 'k')
            ax4.plot(0.0, 0.0, 'r.')
            ax4.axis("equal")

    @QmQkitWrapper.measure
    def quantum_jump_hist_test(self, n_measure, ref_pos, dist, t_wait=25, plot=True):
        """
        The regions defined by the variance must not overlap
        """

        pos1x = ref_pos[0, 0].item()
        pos1y = ref_pos[0, 1].item()
        pos2x = ref_pos[1, 0].item()
        pos2y = ref_pos[1, 1].item()

        # var1 = dist[0].item()
        # var2 = dist[1].item()
        sigma1 = dist[0].item()
        sigma2 = dist[1].item()

        with program() as quantumJumps:

            N = declare(int)
            Is = declare(fixed)
            Qs = declare(fixed)
            Ir = declare(fixed)
            Qr = declare(fixed)
            I = declare(fixed)
            Q = declare(fixed)

            d1 = declare(fixed)
            d2 = declare(fixed)
            state = declare(int, value=-1)  # latch filter : value = -1 (ground state)
            # state = declare(int, value=0) # not latch filter: value = 0,
            count = declare(int, value=0)

            with for_(N, 0, N < n_measure, N + 1):
                # measurement block:
                measure("readout_sig", "signal", None, ("integW_Is", Is), ("integW_Qs", Qs))
                measure("readout_ref", "reference", None, ("integW_Ir", Ir), ("integW_Qr", Qr))

                wait(t_wait, "signal")

                # calculations:
                assign(I, (Is << bit_shift) * (Ir << bit_shift) + (Qs << bit_shift) * (Qr << bit_shift))
                assign(Q, (Is << bit_shift) * (Qr << bit_shift) - (Qs << bit_shift) * (Ir << bit_shift))

                # assign(d1, (I - pos1x) * (I - pos1x) + (Q - pos1y) * (Q - pos1y))
                # assign(d2, (I - pos2x) * (I - pos2x) + (Q - pos2y) * (Q - pos2y))

                assign(d1, pos1x - I)
                assign(d2, I - pos2x)

                # latch filter
                with if_(state == -1):
                    with if_(d2 > sigma2):
                        assign(count, count + 1)
                    with else_():
                        save(count, "hist_down")
                        assign(count, 0)  # start counting from zero
                        assign(state, 1)

                with else_():
                    with if_(d1 > sigma1):
                        assign(count, count + 1)
                    with else_():
                        save(count, "hist_up")
                        assign(count, 0)  # start counting from zero
                        assign(state, -1)

                # no latch filter
                # with if_(d1 < sigma1):
                #     with if_(state == -1):
                #         assign(count, count + 1)
                #     with else_():
                #         with if_(state == 1):
                #             save(count, "hist_up")
                #         assign(count, 0)  # start counting from zero
                #         assign(state, -1)
                # with else_():
                #     with if_(d2 < sigma2):
                #         with if_(state == 1):
                #             assign(count, count + 1)
                #         with else_():
                #             with if_(state == -1):
                #                 save(count, "hist_down")
                #             assign(count, 0)  # start counting from zero
                #             assign(state, 1)
                #     with else_():
                #         with if_(state == 1):
                #             save(count, "hist_up")
                #         with if_(state == -1):
                #             save(count, "hist_down")
                #         assign(state, 0)

                # save to client:
                save(I, "I")
                save(Q, "Q")
                save(d1, "d1")
                save(d2, "d2")
                save(state, "state")
                save(count, "count")

        self.program = quantumJumps
        job = self.qm_config.qm.execute(self.program, duration_limit=0, data_limit=0)
        job.wait_for_all_results()
        res = job.get_results()

        self.program = quantumJumps
        self.I = res.variable_results.I.values
        self.Q = res.variable_results.Q.values
        d1 = res.variable_results.d1.values
        d2 = res.variable_results.d2.values
        self.state = res.variable_results.state.values
        self.t = res.variable_results.state.ts_nsec
        count = res.variable_results.count.values
        hist_up = res.variable_results.hist_up.values
        hist_down = res.variable_results.hist_down.values

        hist_up, _ = np.histogram(hist_up[1:-1], bins=np.arange(hist_up.max() + 2) - 0.5)   # Throw out first and last jump
        hist_down, _ = np.histogram(hist_down[1:-1], bins=np.arange(hist_down.max() + 2) - 0.5)  # Throw out first and last jump

        # qkit data storage
        self.exp_name = "quantum_jumps_{}us".format(4 * t_wait / 1000).replace('.', 'p')
        self.coords = {"t": [self.t / 1000, "us"]}
        self.values = {"state": [state, ["t"], "sate"],
                       "I": [self.I, ["t"], "V"], "Q": [self.Q, ["t"], "V"]}

        if plot:
            self.fig.clear()
            grid = self.fig.add_gridspec(ncols=1, nrows=6, height_ratios=[1, 1, 1, 2, 1, 1])

            ax1 = self.fig.add_subplot(grid[0, 0])
            #ax1.plot(np.arange(n_measure), self.I)
            #ax1.plot(np.arange(n_measure), self.Q)
            ax1.plot(np.arange(n_measure), d1)
            ax1.plot(np.arange(n_measure), d2)

            ax2 = self.fig.add_subplot(grid[1, 0])
            ax2.plot(np.arange(n_measure), self.state)

            ax3 = self.fig.add_subplot(grid[2, 0])
            ax3.plot(np.arange(n_measure), count)

            ax4 = self.fig.add_subplot(grid[3, 0])
            ax4.scatter(self.I, self.Q, 0.1)
            # s = np.linspace(0, 2 * np.pi)
            # ax4.plot(ref_pos[0, 0] + np.sqrt(var) * np.cos(s), ref_pos[0, 1] + np.sqrt(var) * np.sin(s), 'k')
            # ax4.plot(ref_pos[1, 0] + np.sqrt(var) * np.cos(s), ref_pos[1, 1] + np.sqrt(var) * np.sin(s), 'k')
            ax4.plot(0.0, 0.0, 'r.')
            ax4.axvline(pos1x - sigma1, color='C0')
            ax4.axvline(pos2x + sigma2, color='C1')
            ax4.axis("equal")

            ax5 = self.fig.add_subplot(grid[4, 0])
            ax5.bar(np.arange(hist_down.size), hist_down, color="C0", width=1.0, alpha=0.8)
            ax5.set_yscale("symlog")

            ax6 = self.fig.add_subplot(grid[5, 0])
            ax6.bar(np.arange(hist_up.size), hist_up, color="C1", width=1.0, alpha=0.8)
            ax6.set_yscale("symlog")

    @QmQkitWrapper.measure
    def quantum_jump_hist(self, n_measure, ref_pos, dist, t_wait=25, plot=True):
        """
        The regions defined by the variance must not overlap
        """

        pos1x = ref_pos[0, 0].item()
        pos1y = ref_pos[0, 1].item()
        pos2x = ref_pos[1, 0].item()
        pos2y = ref_pos[1, 1].item()

        # var1 = dist[0].item()
        # var2 = dist[1].item()
        sigma1 = dist[0].item()
        sigma2 = dist[1].item()

        with program() as quantumJumps:

            N = declare(int)
            Is = declare(fixed)
            Qs = declare(fixed)
            Ir = declare(fixed)
            Qr = declare(fixed)
            I = declare(fixed)
            Q = declare(fixed)

            d1 = declare(fixed)
            d2 = declare(fixed)
            state = declare(int, value=0)
            count = declare(int)

            with for_(N, 0, N < n_measure, N + 1):
                # measurement block:
                measure("readout_sig", "signal", None, ("integW_Is", Is), ("integW_Qs", Qs))
                measure("readout_ref", "reference", None, ("integW_Ir", Ir), ("integW_Qr", Qr))

                wait(t_wait, "signal")

                # calculations:
                assign(I, (Is << bit_shift) * (Ir << bit_shift) + (Qs << bit_shift) * (Qr << bit_shift))
                assign(Q, (Is << bit_shift) * (Qr << bit_shift) - (Qs << bit_shift) * (Ir << bit_shift))

                # assign(d1, (I - pos1x) * (I - pos1x) + (Q - pos1y) * (Q - pos1y))
                # assign(d2, (I - pos2x) * (I - pos2x) + (Q - pos2y) * (Q - pos2y))

                assign(d1, pos1x - I)
                assign(d2, I - pos2x)

                # latch filter
                with if_(state == -1):
                    with if_(d2 > sigma2):
                        assign(count, count + 1)
                    with else_():
                        save(count, "hist_down")
                        assign(count, 0)  # start counting from zero
                        assign(state, 1)
                with else_():
                    with if_(state == 1):
                        with if_(d1 > sigma1):
                            assign(count, count + 1)
                        with else_():
                            save(count, "hist_up")
                            assign(state, -1)
                            assign(count, 0)  # start counting from zero
                    with else_():  # initialize here to have the first jump length correct
                        with if_(d1 < sigma1):
                            assign(state, -1)
                            assign(count, 0)
                        with else_():
                            with if_(d2 < sigma2):
                                assign(state, 1)
                                assign(count, 0)

                # no latch filter
                # with if_(d1 < sigma1):
                #     with if_(state == -1):
                #         assign(count, count + 1)
                #     with else_():
                #         with if_(state == 1):
                #             save(count, "hist_up")
                #         assign(count, 0)  # start counting from zero
                #         assign(state, -1)
                # with else_():
                #     with if_(d2 < sigma2):
                #         with if_(state == 1):
                #             assign(count, count + 1)
                #         with else_():
                #             with if_(state == -1):
                #                 save(count, "hist_down")
                #             assign(count, 0)  # start counting from zero
                #             assign(state, 1)
                #     with else_():
                #         with if_(state == 1):
                #             save(count, "hist_up")
                #         with if_(state == -1):
                #             save(count, "hist_down")
                #         assign(state, 0)

                # save to client:
                save(I, "I")
                save(Q, "Q")
                save(state, "state")


        self.program = quantumJumps
        job = self.qm_config.qm.execute(self.program, duration_limit=0, data_limit=0)
        job.wait_for_all_results()
        res = job.get_results()

        self.I = res.variable_results.I.values
        self.Q = res.variable_results.Q.values
        self.state = res.variable_results.state.values
        self.t = res.variable_results.state.ts_nsec
        hist_up = res.variable_results.hist_up.values
        hist_down = res.variable_results.hist_down.values

        hist_up, _ = np.histogram(hist_up[:-1], bins=np.arange(hist_up.max() + 1) + 0.5)
        hist_down, _ = np.histogram(hist_down[:-1], bins=np.arange(hist_down.max() + 1) + 0.5)

        # qkit data storage
        self.exp_name = "quantum_jump_hist_{}us".format(4 * t_wait / 1000).replace('.', 'p')
        self.coords = {"t": [self.t / 1000, "us"],
                       "hist_up": [hist_up, "1"],
                       "hist_down": [hist_down, "1"]}   # "hist_down_range": [np.arange(hist_down.size) + 1, "1"]
        self.values = {"state": [self.state, ["t"], "state"],
                       "I": [self.I, ["t"], "V"], "Q": [self.Q, ["t"], "V"]}

        if plot:
            t_rep = np.median(self.t[1:] - self.t[:-1]) / 1000

            self.fig.clear()
            grid = self.fig.add_gridspec(ncols=1, nrows=3, height_ratios=[2, 1, 1])

            ax1 = self.fig.add_subplot(grid[0, 0])
            ax1.scatter(self.I, self.Q, 0.1)
            #s = np.linspace(0, 2 * np.pi)
            #ax1.plot(ref_pos[0, 0] + np.sqrt(var) * np.cos(s), ref_pos[0, 1] + np.sqrt(var) * np.sin(s), 'k')
            #ax1.plot(ref_pos[1, 0] + np.sqrt(var) * np.cos(s), ref_pos[1, 1] + np.sqrt(var) * np.sin(s), 'k')
            ax1.plot(0.0, 0.0, 'r.')
            ax1.axvline(pos1x - sigma1, color='C0')
            ax1.axvline(pos2x + sigma2, color='C1')
            ax1.axis("equal")

            # Maximum likelihood fits and histogram plots
            m = np.sum(hist_down)
            n = np.sum(hist_down * np.arange(hist_down.size))
            p = n / (m + n)
            t_up = - t_rep / np.log(p)
            p_down = m * (1 - p) * p ** np.arange(hist_down.size)

            ax2 = self.fig.add_subplot(grid[1, 0])
            ax2.text(0.9, 0.9, r'$p$ = {:.6f}'.format(p) + "\n" + r'$t_\uparrow$= {:.6f}'.format(t_up),
                     horizontalalignment='right', verticalalignment='top', transform=ax2.transAxes)
            ax2.bar(np.arange(hist_down.size), hist_down, color="C0", width=1.0, alpha=0.8)
            ax2.plot(np.arange(hist_down.size), p_down, color="C0")
            ax2.set_yscale("symlog")

            m = np.sum(hist_up)
            n = np.sum(hist_up * np.arange(hist_up.size))
            p = n / (m + n)
            t_down = - t_rep / np.log(p)
            p_up = m * (1 - p) * p ** np.arange(hist_up.size)

            ax3 = self.fig.add_subplot(grid[2, 0])
            ax3.text(0.9, 0.9, r'$p$ = {:.6f}'.format(p) + "\n" + r'$t_\downarrow$ = {:.6f}'.format(t_down),
                     horizontalalignment='right', verticalalignment='top', transform=ax3.transAxes)
            ax3.bar(np.arange(hist_up.size), hist_up, color="C1", width=1.0, alpha=0.8)
            ax3.plot(np.arange(hist_up.size), p_up, color="C1")
            ax3.set_yscale("symlog")

    @QmQkitWrapper.measure
    def quantum_jump_hist_stab(self, n_measure, n_stab, ref_pos, dist, t_wait=25,
                               stab_state=True, start_state=False, plot=True):
        """ Workaround
        The regions defined by the variance must not overlap
        """
        pos1x = ref_pos[0, 0].item()
        pos1y = ref_pos[0, 1].item()
        pos2x = ref_pos[1, 0].item()
        pos2y = ref_pos[1, 1].item()

        # var1 = dist[0].item()
        # var2 = dist[1].item()
        sigma1 = dist[0].item()
        sigma2 = dist[1].item()

        threshold = (pos1x + pos2x) / 2

        if start_state:
            init_state = 1
        else:
            init_state = -1

        with program() as quantumJumps:

            N = declare(int)
            Is = declare(fixed)
            Qs = declare(fixed)
            Ir = declare(fixed)
            Qr = declare(fixed)
            I = declare(fixed)
            Q = declare(fixed)

            d1 = declare(fixed)
            d2 = declare(fixed)
            state = declare(int, value=init_state)
            count = declare(int, value=0)

            with for_(N, 0, N < n_measure, N + 1):

                with if_(N < n_stab):

                    wait(t_wait, "qubit")  # wait between measurements

                    # measurement block:
                    align("qubit", "signal", "reference")
                    measure("readout_sig", "signal", None, ("integW_Is", Is), ("integW_Qs", Qs))
                    measure("readout_ref", "reference", None, ("integW_Ir", Ir), ("integW_Qr", Qr))

                    # calculations:
                    assign(I, (Is << bit_shift) * (Ir << bit_shift) + (Qs << bit_shift) * (Qr << bit_shift))

                    # wait ring down resonator TODO???

                    align("qubit", "signal", "reference")
                    if stab_state:
                        # excited
                        play("pi_pulse", "qubit", condition=I > threshold)  # excited
                        with if_(I > threshold):
                            save(N, "flip")
                        with else_():
                            wait(12, "qubit")  # qubit pulse length

                    else:
                        play("pi_pulse", "qubit", condition=I < threshold)
                        with if_(I < threshold):
                            save(N, "flip")
                        with else_():
                            wait(12, "qubit")  # qubit pulse length

                with else_():
                    with if_(N == n_stab):
                        if (start_state != stab_state):
                            play("pi_pulse", "qubit")
                        else:
                            wait(12, "qubit")  # qubit pulse length
                    with else_():

                        wait(t_wait, "signal")

                        # measurement block:
                        align("qubit", "signal", "reference")
                        measure("readout_sig", "signal", None, ("integW_Is", Is), ("integW_Qs", Qs))
                        measure("readout_ref", "reference", None, ("integW_Ir", Ir), ("integW_Qr", Qr))

                        # calculations:
                        assign(I, (Is << bit_shift) * (Ir << bit_shift) + (Qs << bit_shift) * (Qr << bit_shift))
                        assign(Q, (Is << bit_shift) * (Qr << bit_shift) - (Qs << bit_shift) * (Ir << bit_shift))

                        # assign(d1, (I - pos1x) * (I - pos1x) + (Q - pos1y) * (Q - pos1y))
                        # assign(d2, (I - pos2x) * (I - pos2x) + (Q - pos2y) * (Q - pos2y))

                        assign(d1, pos1x - I)
                        assign(d2, I - pos2x)

                        # latch filter
                        with if_(state == -1):
                            with if_(d2 > sigma2):
                                assign(count, count + 1)
                            with else_():
                                save(count, "hist_down")
                                assign(count, 0)  # start counting from zero
                                assign(state, 1)
                        with else_():
                            with if_(d1 > sigma1):
                                assign(count, count + 1)
                            with else_():
                                save(count, "hist_up")
                                assign(state, -1)
                                assign(count, 0)  # start counting from zero

                        # save to client:
                        save(I, "I")
                        save(Q, "Q")
                        save(state, "state")

        self.program = quantumJumps
        job = self.qm_config.qm.execute(self.program, duration_limit=0, data_limit=0)
        job.wait_for_all_results()
        res = job.get_results()

        self.I = res.variable_results.I.values
        self.Q = res.variable_results.Q.values
        self.state = res.variable_results.state.values
        self.t = res.variable_results.state.ts_nsec
        hist_up = res.variable_results.hist_up.values
        hist_down = res.variable_results.hist_down.values

        hist_up, _ = np.histogram(hist_up[:-1], bins=np.arange(hist_up.max() + 1) + 0.5)
        hist_down, _ = np.histogram(hist_down[:-1], bins=np.arange(hist_down.max() + 1) + 0.5)

        flip = res.variable_results.flip.values
        t_flip = res.variable_results.flip.ts_nsec
        flip_hist, _ = np.histogram(flip, bins=np.arange(n_stab + 1) - 0.5)

        # qkit data storage
        self.exp_name = "quantum_jump_hist_stab_{}_{}_{}_{}us".format(n_stab, stab_state, start_state, 4 * t_wait / 1000).replace('.', 'p')
        self.coords = {"t": [self.t / 1000, "us"],
                       "hist_up": [hist_up, "1"],
                       "hist_down": [hist_down, "1"],
                       "flip_index": [np.arange(n_stab), "#"],
                       "flip": [flip, "1"],
                       "t_flip": [t_flip / 1000, "ns"]}
        self.values = {"state": [self.state, ["t"], "state"],
                       "I": [self.I, ["t"], "V"], "Q": [self.Q, ["t"], "V"],
                       "flip_hist": [flip_hist, ["flip_index"], "1"]}

        if plot:
            t_rep = np.median(self.t[1:] - self.t[:-1]) / 1000

            self.fig.clear()
            grid = self.fig.add_gridspec(ncols=1, nrows=3, height_ratios=[2, 1, 1])

            ax1 = self.fig.add_subplot(grid[0, 0])
            ax1.scatter(self.I, self.Q, 0.1)
            # s = np.linspace(0, 2 * np.pi)
            # ax1.plot(ref_pos[0, 0] + np.sqrt(var) * np.cos(s), ref_pos[0, 1] + np.sqrt(var) * np.sin(s), 'k')
            # ax1.plot(ref_pos[1, 0] + np.sqrt(var) * np.cos(s), ref_pos[1, 1] + np.sqrt(var) * np.sin(s), 'k')
            ax1.plot(0.0, 0.0, 'r.')
            ax1.axvline(pos1x - sigma1, color='C0')
            ax1.axvline(pos2x + sigma2, color='C1')
            ax1.axis("equal")

            # Maximum likelihood fits and histogram plots
            m = np.sum(hist_down)
            n = np.sum(hist_down * np.arange(hist_down.size))
            p = n / (m + n)
            t_up = - t_rep / np.log(p)
            p_down = m * (1 - p) * p ** np.arange(hist_down.size)

            ax2 = self.fig.add_subplot(grid[1, 0])
            ax2.text(0.9, 0.9, r'$p$ = {:.6f}'.format(p) + "\n" + r'$t_\uparrow$= {:.6f}'.format(t_up),
                     horizontalalignment='right', verticalalignment='top', transform=ax2.transAxes)
            ax2.bar(np.arange(hist_down.size), hist_down, color="C0", width=1.0, alpha=0.8)
            ax2.plot(np.arange(hist_down.size), p_down, color="C0")
            ax2.set_yscale("symlog")

            m = np.sum(hist_up)
            n = np.sum(hist_up * np.arange(hist_up.size))
            p = n / (m + n)
            t_down = - t_rep / np.log(p)
            p_up = m * (1 - p) * p ** np.arange(hist_up.size)

            ax3 = self.fig.add_subplot(grid[2, 0])
            ax3.text(0.9, 0.9, r'$p$ = {:.6f}'.format(p) + "\n" + r'$t_\downarrow$ = {:.6f}'.format(t_down),
                     horizontalalignment='right', verticalalignment='top', transform=ax3.transAxes)
            ax3.bar(np.arange(hist_up.size), hist_up, color="C1", width=1.0, alpha=0.8)
            ax3.plot(np.arange(hist_up.size), p_up, color="C1")
            ax3.set_yscale("symlog")

    @QmQkitWrapper.measure
    def quantum_jump_hist_average(self, n_measure, ref_pos, dist, t_wait=25, avg=10, plot=True):
        """
        The regions defined by the variance must not overlap
        """

        pos1x = ref_pos[0, 0].item()
        pos1y = ref_pos[0, 1].item()
        pos2x = ref_pos[1, 0].item()
        pos2y = ref_pos[1, 1].item()

        # var1 = dist[0].item()
        # var2 = dist[1].item()
        sigma1 = dist[0].item()
        sigma2 = dist[1].item()

        with program() as quantumJumps:

            N = declare(int)
            Is = declare(fixed)
            Qs = declare(fixed)
            Ir = declare(fixed)
            Qr = declare(fixed)
            I = declare(fixed)
            Q = declare(fixed)

            d1 = declare(fixed)
            d2 = declare(fixed)
            state = declare(int)  # latch filter : value = -1 (ground state)
            count = declare(int)
            i = declare(int)
            iter_done = declare(int, value=-1)

            with for_(i, 0, i < avg, i + 1):

                assign(state, 0)  # needed for initialization

                wait(4, "qubit")  # 60ms maximum

                # qubit operation
                # ...

                # align("qubit", "signal", "reference")
                with for_(N, 0, N < n_measure, N + 1):

                    # measurement block:
                    measure("readout_sig", "signal", None, ("integW_Is", Is), ("integW_Qs", Qs))
                    measure("readout_ref", "reference", None, ("integW_Ir", Ir), ("integW_Qr", Qr))

                    wait(t_wait, "signal")

                    # calculations:
                    assign(I, (Is << bit_shift) * (Ir << bit_shift) + (Qs << bit_shift) * (Qr << bit_shift))
                    assign(Q, (Is << bit_shift) * (Qr << bit_shift) - (Qs << bit_shift) * (Ir << bit_shift))

                    # assign(d1, (I - pos1x) * (I - pos1x) + (Q - pos1y) * (Q - pos1y))
                    # assign(d2, (I - pos2x) * (I - pos2x) + (Q - pos2y) * (Q - pos2y))

                    assign(d1, pos1x - I)
                    assign(d2, I - pos2x)

                    # latch filter
                    with if_(state == -1):
                        with if_(d2 > sigma2):
                            assign(count, count + 1)
                        with else_():
                            save(count, "hist_down")
                            assign(state, 1)
                            assign(count, 0)  # start counting from zero
                    with else_():
                        with if_(state == 1):
                            with if_(d1 > sigma1):
                                assign(count, count + 1)
                            with else_():
                                save(count, "hist_up")
                                assign(state, -1)
                                assign(count, 0)  # start counting from zero
                        with else_():  # initialize here to have the first jump length correct
                            with if_(d1 < sigma1):
                                assign(state, -1)
                                assign(count, 0)
                            with else_():
                                with if_(d2 < sigma2):
                                    assign(state, 1)
                                    assign(count, 0)

                    # save to client:
                    save(I, "I")
                    save(Q, "Q")
                    save(state, "state")

                save(iter_done, "hist_down")
                save(iter_done, "hist_up")

        self.program = quantumJumps
        job = self.qm_config.qm.execute(self.program, duration_limit=0, data_limit=0)
        job.wait_for_all_results()
        res = job.get_results()

        self.I = res.variable_results.I.values
        self.Q = res.variable_results.Q.values
        self.state = res.variable_results.state.values
        self.t = res.variable_results.state.ts_nsec
        hist_up = res.variable_results.hist_up.values
        hist_down = res.variable_results.hist_down.values

        hist_up, _ = np.histogram(hist_up[1:-1], bins=np.arange(hist_up.max() + 1) + 0.5)
        hist_down, _ = np.histogram(hist_down[1:-1], bins=np.arange(hist_down.max() + 1) + 0.5)

        # qkit data storage
        self.exp_name = "quantum_jump_hist_{}us".format(4 * t_wait / 1000).replace('.', 'p')
        self.coords = {"t": [self.t / 1000, "us"],
                       "hist_up": [hist_up, "1"],
                       "hist_down": [hist_down, "1"]}  # "hist_down_range": [np.arange(hist_down.size) + 1, "1"]
        self.values = {"state": [self.state, ["t"], "state"],
                       "I": [self.I, ["t"], "V"], "Q": [self.Q, ["t"], "V"]}

        if plot:

            t_rep = np.median(self.t[1:] - self.t[:-1]) / 1000

            self.fig.clear()
            grid = self.fig.add_gridspec(ncols=1, nrows=3, height_ratios=[2, 1, 1])

            self.ax1 = self.fig.add_subplot(grid[0, 0])
            self.ax1.scatter(self.I, self.Q, 0.1)
            # s = np.linspace(0, 2 * np.pi)
            # self.ax1.plot(ref_pos[0, 0] + np.sqrt(var) * np.cos(s), ref_pos[0, 1] + np.sqrt(var) * np.sin(s), 'k')
            # self.ax1.plot(ref_pos[1, 0] + np.sqrt(var) * np.cos(s), ref_pos[1, 1] + np.sqrt(var) * np.sin(s), 'k')
            self.ax1.plot(0.0, 0.0, 'r.')
            self.ax1.axvline(pos1x - sigma1, color='C0')
            self.ax1.axvline(pos2x + sigma2, color='C1')
            self.ax1.axis("equal")

            # Maximum likelihood fits and histogram plots
            m = np.sum(hist_down)
            n = np.sum(hist_down * np.arange(hist_down.size))
            p = n / (m + n)
            t_up = - t_rep / np.log(p)
            p_down = m * (1 - p) * p ** np.arange(hist_down.size)

            ax2 = self.fig.add_subplot(grid[1, 0])
            ax2.text(0.9, 0.9, r'$p$ = {:.6f}'.format(p) + "\n" + r'$t_\uparrow$= {:.6f}'.format(t_up),
                     horizontalalignment='right', verticalalignment='top', transform=ax2.transAxes)
            ax2.bar(np.arange(hist_down.size), hist_down, color="C0", width=1.0, alpha=0.8)
            ax2.plot(np.arange(hist_down.size), p_down, color="C0")
            ax2.set_yscale("symlog")

            m = np.sum(hist_up)
            n = np.sum(hist_up * np.arange(hist_up.size))
            p = n / (m + n)
            t_down = - t_rep / np.log(p)
            p_up = m * (1 - p) * p ** np.arange(hist_up.size)

            ax3 = self.fig.add_subplot(grid[2, 0])
            ax3.text(0.9, 0.9, r'$p$ = {:.6f}'.format(p) + "\n" + r'$t_\downarrow$ = {:.6f}'.format(t_down),
                     horizontalalignment='right', verticalalignment='top', transform=ax3.transAxes)
            ax3.bar(np.arange(hist_up.size), hist_up, color="C1", width=1.0, alpha=0.8)
            ax3.plot(np.arange(hist_up.size), p_up, color="C1")
            ax3.set_yscale("symlog")

    @QmQkitWrapper.measure
    def quantum_jump_hist_stab_avg(self, n_measure,  n_stab, ref_pos, dist, t_stab=200, t_measure=200,
                                   stab_state=False, start_state=False, avg=100, plot=True):
        """
        The regions defined by the variance must not overlap
        """

        pos1x = ref_pos[0, 0].item()
        pos1y = ref_pos[0, 1].item()
        pos2x = ref_pos[1, 0].item()
        pos2y = ref_pos[1, 1].item()

        sigma1 = dist[0].item()
        sigma2 = dist[1].item()

        threshold = (pos1x + pos2x) / 2
        print(threshold)

        with program() as quantumJumps:

            n = declare(int)
            i = declare(int)
            j = declare(int)
            iter_done = declare(int, value=-1)

            Is = declare(fixed)
            Qs = declare(fixed)
            Ir = declare(fixed)
            Qr = declare(fixed)
            I = declare(fixed)
            Q = declare(fixed)

            d1 = declare(fixed)
            d2 = declare(fixed)
            state = declare(int)
            count = declare(int)

            with for_(i, 0, i < avg, i + 1):

                save(iter_done, "flip")
                save(iter_done, "hist_down")
                save(iter_done, "hist_up")

                assign(state, 0)  # initialization

                # measurement block:
                measure("readout_sig", "signal", None, ("integW_Is", Is), ("integW_Qs", Qs))
                measure("readout_ref", "reference", None, ("integW_Ir", Ir), ("integW_Qr", Qr))

                # calculations:
                assign(I, (Is << bit_shift) * (Ir << bit_shift) + (Qs << bit_shift) * (Qr << bit_shift))

                with for_(j, 0, j < n_stab, j + 1):  # stabilization for n_stab

                    # wait ring down resonator to have a more precise pi pulse???

                    if stab_state:
                        # excited
                        play("pi_pulse", "qubit", condition=I > threshold)  # excited
                        with if_(I > threshold):
                            save(j, "flip")
                        with else_():
                            wait(12, "qubit")  # qubit pulse length

                    else:
                        play("pi_pulse", "qubit", condition=I < threshold)
                        with if_(I < threshold):
                            save(j, "flip")
                        with else_():
                            wait(12, "qubit")  # qubit pulse length

                    wait(t_stab, "qubit")  # wait between measurements

                    # measurement block:
                    align("qubit", "signal", "reference")
                    measure("readout_sig", "signal", None, ("integW_Is", Is), ("integW_Qs", Qs))
                    measure("readout_ref", "reference", None, ("integW_Ir", Ir), ("integW_Qr", Qr))

                    # calculations:
                    assign(I, (Is << bit_shift) * (Ir << bit_shift) + (Qs << bit_shift) * (Qr << bit_shift))

                align("qubit", "signal", "reference")
                if start_state:
                    # excited
                    play("pi_pulse", "qubit", condition=I > threshold)  # excited
                    with if_(I > threshold):
                        save(j, "flip")
                    with else_():
                        wait(12, "qubit")  # qubit pulse length

                else:
                    play("pi_pulse", "qubit", condition=I < threshold)
                    with if_(I < threshold):
                        save(j, "flip")
                    with else_():
                        wait(12, "qubit")  # qubit pulse length

                align("qubit", "signal", "reference")
                with for_(n, 0, n < n_measure, n + 1):

                    # measurement block:
                    measure("readout_sig", "signal", None, ("integW_Is", Is), ("integW_Qs", Qs))
                    measure("readout_ref", "reference", None, ("integW_Ir", Ir), ("integW_Qr", Qr))

                    wait(t_measure, "signal")

                    # calculations:
                    assign(I, (Is << bit_shift) * (Ir << bit_shift) + (Qs << bit_shift) * (Qr << bit_shift))
                    assign(Q, (Is << bit_shift) * (Qr << bit_shift) - (Qs << bit_shift) * (Ir << bit_shift))

                    # assign(d1, (I - pos1x) * (I - pos1x) + (Q - pos1y) * (Q - pos1y))
                    # assign(d2, (I - pos2x) * (I - pos2x) + (Q - pos2y) * (Q - pos2y))

                    assign(d1, pos1x - I)
                    assign(d2, I - pos2x)

                    # latch filter
                    with if_(state == -1):
                        with if_(d2 > sigma2):
                            assign(count, count + 1)
                        with else_():
                            save(count, "hist_down")
                            assign(count, 0)  # start counting from zero
                            assign(state, 1)
                    with else_():
                        with if_(state == 1):
                            with if_(d1 > sigma1):
                                assign(count, count + 1)
                            with else_():
                                save(count, "hist_up")
                                assign(state, -1)
                                assign(count, 0)  # start counting from zero
                        with else_():  # initialize here to have the first jump length correct
                            with if_(d1 < sigma1):
                                assign(state, -1)
                                assign(count, 0)
                            with else_():
                                with if_(d2 < sigma2):
                                    assign(state, 1)
                                    assign(count, 0)

                    # save to client:
                    save(I, "I")
                    save(Q, "Q")
                    save(state, "state")

                align("qubit", "signal", "reference")
                wait(250, "qubit")  # 10ms

        self.program = quantumJumps
        job = self.qm_config.qm.simulate(self.program, SimulationConfig(100000))
        samples = job.get_simulated_samples()

        result = job.result_handles
        j_handle = result.get("flip")
        print(j_handle.fetch_all())

        samples.con1.plot()
        plt.show()


        #if plot:
        #    self.fig.clear()
        #    grid = self.fig.add_gridspec(ncols=1, nrows=3, height_ratios=[2, 1, 1])
        #
        #    ax1 = self.fig.add_subplot(grid[0, 0])
        #    ax1.plot(samples., self.Q, 0.1)
        #    # s = np.linspace(0, 2 * np.pi)
        #    # self.ax1.plot(ref_pos[0, 0] + np.sqrt(var) * np.cos(s), ref_pos[0, 1] + np.sqrt(var) * np.sin(s), 'k')
        #    # self.ax1.plot(ref_pos[1, 0] + np.sqrt(var) * np.cos(s), ref_pos[1, 1] + np.sqrt(var) * np.sin(s), 'k')
        #    ax1.plot(0.0, 0.0, 'r.')


        #job = self.qm_config.qm.execute(self.program, duration_limit=0, data_limit=0)

        # job.wait_for_all_results()
        # res = job.get_results()
        #
        # self.I = np.reshape(res.variable_results.I.values, (avg, n_measure))
        # self.Q = np.reshape(res.variable_results.Q.values, (avg, n_measure))
        # self.state = np.reshape(res.variable_results.state.values, (avg, n_measure))
        # self.t = np.reshape(res.variable_results.state.ts_nsec, (avg, n_measure))
        # self.t -= self.t[:, 0, np.newaxis]
        #
        # list_up = res.variable_results.hist_up.values
        # list_down = res.variable_results.hist_down.values
        # hist_up_avg, _ = np.histogram(list_up[:-1], bins=np.arange(list_up.max() + 1) + 0.5)
        # hist_down_avg, _ = np.histogram(list_down[:-1], bins=np.arange(list_down.max() + 1) + 0.5)
        #
        # flip = res.variable_results.flip.values
        # t_flip = res.variable_results.flip.ts_nsec
        # flip_hist, _ = np.histogram(flip, bins=np.arange(n_stab + 1) - 0.5)
        #
        # # qkit data storage
        # self.exp_name = "quantum_jump_hist_stabilized_{}_{}_{}".format(n_stab, stab_state, start_state)
        # self.coords = {"avg": [np.arange(avg), "#"],
        #                "n_measure": [np.arange(n_measure), "#"],
        #                "list_up": [list_up, "1"],
        #                "list_down": [list_down, "1"],
        #                "hist_up_avg": [hist_up_avg, "1"],
        #                "hist_down_avg": [hist_down_avg, "1"],
        #                "flip_index": [np.arange(n_stab), "#"],
        #                "flip": [flip, "1"],
        #                "t_flip": [t_flip / 1000, "ns"]}
        # self.values = {"state": [self.state, ["avg", "n_measure"], "state"],
        #                "t": [self.t / 1000, ["avg", "n_measure"], "us"],
        #                "I": [self.I, ["avg", "n_measure"], "V"], "Q": [self.Q, ["avg", "n_measure"], "V"],
        #                "flip_hist": [flip_hist, ["flip_index"], "1"]}
        #
        # if plot:
        #
        #     t_rep = np.mean(self.t[:, 1:] - self.t[:, :-1]) / 1000
        #
        #     self.fig.clear()
        #     grid = self.fig.add_gridspec(ncols=1, nrows=3, height_ratios=[2, 1, 1])
        #
        #     ax1 = self.fig.add_subplot(grid[0, 0])
        #     ax1.scatter(self.I, self.Q, 0.1)
        #     # s = np.linspace(0, 2 * np.pi)
        #     # self.ax1.plot(ref_pos[0, 0] + np.sqrt(var) * np.cos(s), ref_pos[0, 1] + np.sqrt(var) * np.sin(s), 'k')
        #     # self.ax1.plot(ref_pos[1, 0] + np.sqrt(var) * np.cos(s), ref_pos[1, 1] + np.sqrt(var) * np.sin(s), 'k')
        #     ax1.plot(0.0, 0.0, 'r.')
        #     ax1.axvline(pos1x - sigma1, color='C0')
        #     ax1.axvline(pos2x + sigma2, color='C1')
        #     ax1.axis("equal")
        #
        #     # Maximum likelihood fit
        #
        #     m = np.sum(hist_down_avg)
        #     n = np.sum(hist_down_avg * np.arange(hist_down_avg.size))
        #     p = n / (m + n)
        #     t_up = - t_rep / np.log(p)
        #     p_down = m * (1 - p) * p ** np.arange(hist_down_avg.size)
        #
        #     ax2 = self.fig.add_subplot(grid[1, 0])
        #     ax2.text(0.9, 0.9, r'$p$ = {:.6f}'.format(p) + "\n" + r'$t_\uparrow$= {:.6f}'.format(t_up),
        #              horizontalalignment='right', verticalalignment='top', transform=ax2.transAxes)
        #     ax2.bar(np.arange(hist_down_avg.size), hist_down_avg, color="C0", width=1.0, alpha=0.8)
        #     ax2.plot(np.arange(hist_down_avg.size), p_down, color="C0")
        #     ax2.set_yscale("symlog")
        #
        #     m = np.sum(hist_up_avg)
        #     n = np.sum(hist_up_avg * np.arange(hist_up_avg.size))
        #     p = n / (m + n)
        #     t_down = - t_rep / np.log(p)
        #     p_up = m * (1 - p) * p ** np.arange(hist_up_avg.size)
        #
        #     ax3 = self.fig.add_subplot(grid[2, 0])
        #     ax3.text(0.9, 0.9, r'$p$ = {:.6f}'.format(p) + "\n" + r'$t_\downarrow$ = {:.6f}'.format(t_down),
        #              horizontalalignment='right', verticalalignment='top', transform=ax3.transAxes)
        #     ax3.bar(np.arange(hist_up_avg.size), hist_up_avg, color="C1", width=1.0, alpha=0.8)
        #     ax3.plot(np.arange(hist_up_avg.size), p_up, color="C1")
        #     ax3.set_yscale("symlog")


import pyqtgraph as pg
from pyqtgraph.Qt import QtCore, QtGui
import numpy as np

class ImageWidget(QtGui.QWidget):
    def __init__(self, parent=None):
        super(ImageWidget, self).__init__(parent)

        # Interpret image data as row-major instead of col-major
        pg.setConfigOptions(imageAxisOrder='row-major')

        pg.mkQApp()
        self.win = pg.GraphicsLayoutWidget()
        self.win.setWindowTitle('pyqtgraph example: Image Analysis')

        # A plot1 area (ViewBox + axes) for displaying the image
        #self.plot1 = self.win.addItem() # addPlot()

        self.vb = pg.ViewBox(enableMouse=True, enableMenu=True)

        # Item for displaying image data
        self.item = pg.ImageItem()
        self.vb.addItem(self.item)
        self.win.addItem(self.vb, row=0, col=1)

        #self.vb.setAspectLocked()
        #self.vb.setRange(xRange=(0, 10), yRange=(0, 10))

        # Item for x-axis
        self.win.nextRow()
        self.xaxis = MyStringAxis(orientation="bottom", linkView=self.vb)
        self.xaxis.setLabel(text="x axis", units="arb.")
        self.yaxis = pg.AxisItem(orientation="left", linkView=self.vb)
        self.yaxis.setLabel(text="y axis", units="arb.")
        self.win.addItem(self.xaxis, row=1, col=1)
        self.win.addItem(self.yaxis, row=0, col=0)


        # Custom ROI for selecting an image region
        self.ROI = pg.ROI([-8, 14], [6, 5])
        self.ROI.addScaleHandle([0.5, 1], [0.5, 0.5])
        self.ROI.addScaleHandle([0, 0.5], [0.5, 0.5])
        self.vb.addItem(self.ROI)
        self.ROI.setZValue(10)  # make sure ROI is drawn above image

        # Isocurve drawing
        self.iso = pg.IsocurveItem(level=0.8, pen='g')
        self.iso.setParentItem(self.item)
        self.iso.setZValue(5)

        # Contrast/color control
        self.hist = pg.HistogramLUTItem()
        self.hist.setImageItem(self.item)
        self.win.addItem(self.hist, row=0, col=2)

        # Draggable line for setting isocurve level
        self.isoLine = pg.InfiniteLine(angle=0, movable=True, pen='g')
        self.hist.vb.addItem(self.isoLine)
        self.hist.vb.setMouseEnabled(y=False) # makes user interaction a little easier
        self.isoLine.setValue(0.8)
        self.isoLine.setZValue(1000) # bring iso line above contrast controls

        # Another plot1 area for displaying ROI data
        self.plot2 = self.win.addPlot(row=2, col=0, colspan=2)
        self.plot2.setMaximumHeight(250)
        self.win.resize(800, 800)
        self.win.show()

        # Generate image self.data
        self.data = np.random.normal(size=(200, 100))
        self.data[20:80, 20:80] += 2.
        self.data = pg.gaussianFilter(self.data, (3, 3))
        self.data += np.random.normal(size=(200, 100)) * 0.1
        self.item.setImage(self.data)
        self.hist.setLevels(self.data.min(), self.data.max())

        # build isocurves from smoothed self.data
        self.iso.setData(pg.gaussianFilter(self.data, (2, 2)))

        # set position and scale of image
        #self.item.scale(0.2, 0.2)
        #self.item.translate(-50, 0)

        ticks = np.arange(self.data.shape[1])
        tickDictMajor = []
        for i, tick in enumerate(ticks):
            tickDictMajor.append(str(i**2))
        self.xaxis.set_tick_dict(dict(enumerate(tickDictMajor)))

        #strings = self.xaxis.tickValues()

        # zoom to fit imageo
        #self.vb.autoRange()

        self.ROI.sigRegionChanged.connect(self.updatePlot)
        self.updatePlot()

        self.isoLine.sigDragged.connect(self.updateIsocurve)

    # Callbacks for handling user interaction
    def updatePlot(self):
        selected = self.ROI.getArrayRegion(self.data, self.item)
        self.plot2.plot(selected.mean(axis=0), clear=True)

    def updateIsocurve(self):
        self.iso.setLevel(self.isoLine.value())


class MyStringAxis(pg.AxisItem):
    def __init__(self, *args, **kwargs):
        pg.AxisItem.__init__(self, *args, **kwargs)
        self.x_values = None
        self.x_strings = None

    def set_tick_dict(self, xdict):

        self.x_values = np.asarray(list(xdict.keys())) + 0.5
        self.x_strings = np.asarray(list(xdict.values()))

    def tickStrings(self, values, scale, spacing):
        strings = []
        for v in values:
            # vs is the original tick value
            vs = v * scale
            # if we have vs in our values, show the string
            # otherwise show nothing
            if vs in self.x_values:
                # Find the string with x_values closest to vs
                vstr = self.x_strings[np.abs(self.x_values - vs).argmin()]
            else:
                vstr = ""
            strings.append(vstr)
        return strings

    def tickSpacing(self, minVal, maxVal, size, offset=0.5):
        """Return values describing the desired spacing and offset of ticks.

        This method is called whenever the axis needs to be redrawn and is a
        good method to override in subclasses that require control over tick locations.

        The return value must be a list of tuples, one for each set of ticks::

            [
                (major tick spacing, offset),
                (minor tick spacing, offset),
                (sub-minor tick spacing, offset),
                ...
            ]
        """
        # First check for override tick spacing
        if self._tickSpacing is not None:
            return self._tickSpacing

        dif = abs(maxVal - minVal)
        if dif == 0:
            return []

        ## decide optimal minor tick spacing in pixels (this is just aesthetics)
        optimalTickCount = max(2., np.log(size))

        ## optimal minor tick spacing
        optimalSpacing = dif / optimalTickCount

        ## the largest power-of-10 spacing which is smaller than optimal
        p10unit = 10 ** np.floor(np.log10(optimalSpacing))

        ## Determine major/minor tick spacings which flank the optimal spacing.
        intervals = np.array([1., 2., 10., 20., 100.]) * p10unit
        minorIndex = 0
        while intervals[minorIndex + 1] <= optimalSpacing:
            minorIndex += 1

        levels = [
            (intervals[minorIndex + 2], offset),
            (intervals[minorIndex + 1], offset),
            # (intervals[minorIndex], offset)    ## Pretty, but eats up CPU
        ]

        if self.style['maxTickLevel'] >= 2:
            ## decide whether to include the last level of ticks
            minSpacing = min(size / 20., 30.)
            maxTickCount = size / minSpacing
            if dif / intervals[minorIndex] <= maxTickCount:
                levels.append((intervals[minorIndex], offset))

        return levels


## Start Qt event loop unless running in interactive mode or using     pyside.
if __name__ == '__main__':

    app = QtGui.QApplication([])
    app.setStyle(QtGui.QStyleFactory.create("Cleanlooks"))

    image_widget = ImageWidget()

    import sys
    if (sys.flags.interactive != 1) or not hasattr(QtCore,     'PYQT_VERSION'):
        QtGui.QApplication.instance().exec_()



import logging
import json
import qkit
#from qkit.measure.measurement_class.Measurement import _JSON_instruments_dict
from qkit.measure.json_handler import QkitJSONEncoder, QkitJSONDecoder
import os

def open_log_file(path,log_level=logging.INFO):
    fn, ext = os.path.splitext(path)
    fn = fn + '.log'
    if len(logging.getLogger().handlers) > 0:
      formatter = logging.getLogger().handlers[0].formatter
    else:
      formatter = None

    log_file_handler = logging.FileHandler(fn)
    log_file_handler.setLevel(log_level)
    log_file_handler.setFormatter(formatter)
    logging.getLogger().addHandler(log_file_handler)
    logging.debug('Added log_file_handler. path="%s", formatter="%s"' % (fn, str(formatter)))
    return log_file_handler

def close_log_file(log_file_handler):
    if log_file_handler is not None:
        logging.getLogger().removeHandler(log_file_handler)
        log_file_handler.close()
        log_file_handler = None

def get_instrument_settings(path):
    fn, ext = os.path.splitext(path)
    instr_dict = {}
    for ins_name in qkit.instruments.get_instruments():
        ins = qkit.instruments.get(ins_name)
        param_dict = {}
        for (param, popts) in _dict_to_ordered_tuples(ins.get_parameters()):
            param_dict.update({param:ins.get(param, query=False, channels=popts)})
            try:
                if popts.get('offset',False):
                    param_dict.update({param+"_offset": ins._offsets[param]})
            except:
                pass
        instr_dict.update({ins_name:param_dict})
    with open(fn+'.set','w+') as filehandler:
        json.dump(obj=instr_dict, fp=filehandler, cls=QkitJSONEncoder, indent = 4, sort_keys=True)

    return instr_dict
    """
    fn_log = fn + '.set'
    f = open(fn_log, 'w+')
    f.write('Filename: %s\n' % fn)
    settings = ''

    inslist = _dict_to_ordered_tuples(qkit.instruments.get_instruments())
    for (iname, ins) in inslist:
        settings += 'Instrument: %s (%s)\n' % (iname, ins.get_type())
        parlist = _dict_to_ordered_tuples(ins.get_parameters())
        for (param, popts) in parlist:
            settings += '\t%s: %s\n' % (param, ins.get(param, query=False, channels=popts))

    f.write(settings)
    f.close()
    return settings
    """

def _dict_to_ordered_tuples(dic):
    '''Convert a dictionary to a list of tuples, sorted by key.'''
    if dic is None:
        return []
    keys = sorted(dic.keys())
    ret = [(key, dic[key]) for key in keys]
    return ret



# Anritsu MS2830A
# Julius Krause <julius.krause@student.kit.edu> 2016
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# Refer to the 'MS2690A/MS2691A/MS2692A or MS2830A Signal
# Analyzer Operation Manual (Mainframe Remote Control)' for detailed
# specifications on IEEE488.2 common device messages and application
# common device messages. 


from qkit.core.instrument_base import Instrument
from qkit import visa
import types
import logging
from time import sleep
import numpy

class Anritsu_MS2830A(Instrument):
    '''
    This is the python driver for the Anritsu MS2830A Signal Source Analyzer
    It only provides very basic functionality

    Usage:
    Initialise with
    <name> = instruments.create('<name>', address='<GPIB address>', reset=<bool>)
    
    '''

    def __init__(self, name, address, comp_mode = False, channel_index = 1):
        '''
        Initializes 

        Input:
            name (string)    : name of the instrument
            address (string) : GPIB address
        '''
        
        logging.info(__name__ + ' : Initializing instrument')
        Instrument.__init__(self, name, tags=['physical'])

        self._address = address
        self._comp_mode = comp_mode
        self._visainstrument = visa.instrument(self._address)

        # Implement parameters
        self.add_parameter('centerfreq', type=float,
            flags=Instrument.FLAG_GETSET,
            minval=0, maxval=26.5e9,
            tags=['sweep'])

        self.add_parameter('nop', type=float,
            flags=Instrument.FLAG_GETSET,
            minval=155, maxval=30001,
            tags=['sweep'])
        self.add_parameter('freqspan', type=float,
            flags=Instrument.FLAG_GETSET,tags=['sweep'])                    

        self.add_parameter('powerunit', type=str,
            flags=Instrument.FLAG_GETSET,tags=['sweep'])

        self.add_parameter('startfreq', type=float,
            flags=Instrument.FLAG_GETSET,tags=['sweep']) 

        self.add_parameter('stopfreq', type=float,
            flags=Instrument.FLAG_GETSET,tags=['sweep']) 

        self.add_parameter('sweeptime', type=float,
            flags=Instrument.FLAG_GET,tags=['sweep'])

        self.add_parameter('resolutionBW', type=float,
            flags=Instrument.FLAG_GETSET,tags=['sweep'])

        self.add_parameter('videoBW', type=float,
            flags=Instrument.FLAG_GETSET,tags=['sweep'])           
            
        # Implement functions
        #self.add_function('set_continuous_sweep_mode')
        #self.add_function('set_freq_center')
        #self.add_function('set_freq_span')
        #self.add_function('set_marker')
        #self.add_function('set_powerunit')
        #self.add_function('get_marker_level')
        #self.add_function('avg_clear')
        #self.add_function('avg_status')
        
        #self._oldspan = self.get_span()
        #self._oldnop = self.get_nop()
        #if self._oldspan==0.002:
        #  self.set_zerospan(True)
        
        #self.get_all()
    
    def get_all(self):        
        self.get_marker_level(1)

        
    ###
    #Communication with device
    ###    
    
    
    ##new below

    def do_set_resolutionBW(self, BW):
        '''
        sets the resolution bandwidth
        '''
        self._visainstrument.write('BAND:BWID %e'%(BW))

    def do_get_resolutionBW(self):
        '''
        gets the resolution bandwidth
        '''
        return float(self._visainstrument.ask('BAND?'))

    def do_set_videoBW(self, BW):
        '''
        sets the video bandwidth
        '''
        self._visainstrument.write('DISP:TRAC:X:WIDT %e'%(BW))

    def do_get_videoBW(self):
        '''
        gets the video bandwidth
        '''
        return float(self._visainstrument.ask('DISP:TRAC:X:WIDT?'))

    def do_get_sweeptime(self):
        '''
        gets the center frequency
        '''
        return float(self._visainstrument.ask('swe:time?'))

    
    def do_set_centerfreq(self, centerfreq):
        '''
        sets the center frequency
        '''
        self._visainstrument.write('freq:cent %e'%(centerfreq))

    def do_get_centerfreq(self):
        '''
        gets the center frequency
        '''
        return float(self._visainstrument.ask('freq:cent?'))

    def do_set_freqspan(self, freqspan):
        '''
        sets the frequency span
        '''
        self._visainstrument.write('freq:span %e'%(freqspan))
    
    
    def do_get_freqspan(self):
        '''
        get the frequency span
        
        '''
        return float(self._visainstrument.ask('freq:span?'))

    def do_set_startfreq(self, freq):
        self._visainstrument.write('freq:star %e'%(freq))

    def do_get_startfreq(self):
        return float(self._visainstrument.ask('freq:star?'))

    def do_set_stopfreq(self, freq):
        self._visainstrument.write('freq:stop %e'%(freq))

    def do_get_stopfreq(self):
        return float(self._visainstrument.ask('freq:stop?'))
        
    def do_set_nop(self, nop):
        self._visainstrument.write('swe:poin %i'%(nop))

    def do_get_nop(self):
        return int(self._visainstrument.ask('swe:poin?'))
        
    def do_set_powerunit(self,unit):
        '''
        sets the unit for powers
        provide unit as a string! ("DBm")
        '''
        self._visainstrument.write('unit:pow %s'%(unit))
        
    def do_get_powerunit(self):
        '''
        gets the power unit for powers
        '''
        return self._visainstrument.ask('unit:pow')
        
        
    def do_get_averages(self):
        '''
        Get number of averages

        Input:
            None
        Output:
            number of averages
        '''
        
        return int(self._visainstrument.ask(':AVER:COUNT?'))

    def do_get_sweeptime_averages(self):
        if self.get_Average():
            return self.get_sweeptime() * self.get_averages()
        else:
            return self.get_sweeptime()        
    
    def set_marker(self,marker,frequency):
        '''
        sets marker number marker to frequency
        
        '''
        self._visainstrument.write('calc:mark%i:x %e'%(marker, frequency))
        self.enable_marker(marker)
        
    def get_marker(self,marker):
        '''
        gets frequency of marker
        
        '''
        return float(self._visainstrument.ask('calc:mark%i:x?'%(marker)))
    
    def get_marker_level(self,marker):
        '''
        gets power level of indicated marker
        
        '''
        return float(self._visainstrument.ask('calc:mark%i:y?'%(marker)))
        
#    def set_continuous_sweep_mode(self,value):
        '''
        value='ON' Continuous sweep
        value='OFF' Single sweep
        '''
        #self._visainstrument.write('INIT:CONT %s'%(value))    
    
    def get_frequencies(self):
        '''
        returns an array with the frequencies of the points returned by get_trace()
        ideally suitable as x-axis for plots
        '''
        return linspace(self.get_startfreq(),self.get_stopfreq(),self.get_nop())
    
#    def enable_marker(self,marker,state='ON'):
        '''
        ON or OFF
        '''
        #self._visainstrument.write('CALC:MARK%i %s'%(marker,state))    
    
#    def sweep(self):
        '''
        perform a sweep and wait for it to finish
        '''
        #self._visainstrument.write('INIT; *WAI')    
    
    def get_trace(self, tracenumber=1):
        tmp = self._visainstrument.ask_for_values(':trac:data? trace%i'%tracenumber)
        if self._comp_mode:
            tmp = [tmp,numpy.zeros_like(tmp)]
        return tmp
    
    def get_frequencies(self):
        '''
        returns an array with the frequencies of the points returned by get_trace()
        ideally suitable as x-axis for plots
        '''
        return numpy.linspace(self.get_startfreq(),self.get_stopfreq(),self.get_nop())

    
    ##enable for debugging and testing new functions:

    def write(self, command):
        self._visainstrument.write(command)    
    
    def ask(self,command):
        return self._visainstrument.ask(command)    

